r"""Mutation models."""

import gctree.utils as utils

from ete3 import TreeNode
import numpy as np
from scipy.stats import poisson
import random
from Bio.Seq import Seq
import historydag as hdag
from multiset import FrozenMultiset
from typing import Tuple, List, Callable, Optional
import itertools
import math


class MutationModel:
    r"""A class for a mutation model, and functions to mutate sequences.

    Args:
        mutability_file: S5F format mutabilities
        substitution_file: S5F format substitution biases
        mutation_order: whether or not to mutate sequences using a context sensitive manner
                        where mutation order matters
        with_replacement: allow the same position to mutate multiple times on a single branch

    Notes:
        ``mutability_file`` shall be a csv file with the first column containing fivemers,
        and the second column containing mutability scores.
        An example can be found at https://bitbucket.org/kleinstein/shazam/src/master/data-raw/HS5F_Mutability.csv

        For example:

        .. code-block:: text

            Fivemer,Mutability,...
            TCGGG,0.03542,...
            GCCGG,0.02241675,...
            GCCGC,0.06789,...
            .
            .
            .


        ``substitution_file`` shall be a csv file with the first column containing fivemers,
        and the next four columns containing targeting probabilities for bases A, C, G,
        and T, respectively.
        An example can be found at https://bitbucket.org/kleinstein/shazam/src/master/data-raw/HS5F_Substitution.csv

        For example:

        .. code-block:: text

            Fivemer,A,C,G,T,...
            AAAAA,0,0.33,0.33,0.34,...
            AAAAC,0,0.5000,0.2500,0.2500,...
            AAAAG,0,0.65,0.15,0.20,...
            .
            .
            .
    """

    def __init__(
        self,
        mutability_file: Optional[str] = None,
        substitution_file: Optional[str] = None,
        mutation_order: bool = True,
        with_replacement: bool = True,
    ):
        self.mutation_order = mutation_order
        self.with_replacement = with_replacement
        if mutability_file is not None and substitution_file is not None:
            self.context_model = {}
            with open(mutability_file, "r") as f:
                # eat header
                f.readline()
                for line in f:
                    if line[0] == '"':
                        motif, score = line.replace('"', "").split()[:2]
                    else:
                        motif, score = line.replace(",", " ").split()[:2]
                    self.context_model[motif] = float(score)

            # kmer k
            self.k = None
            with open(substitution_file, "r") as f:
                # eat header
                f.readline()
                for line in f:
                    if line[0] == '"':
                        fields = line.replace('"', "").split()
                    else:
                        fields = line.replace(",", " ").split()
                    motif = fields[0]
                    if self.k is None:
                        self.k = len(motif)
                        assert self.k % 2 == 1
                    else:
                        assert len(motif) == self.k
                    self.context_model[motif] = (
                        self.context_model[motif],
                        {b: float(x) for b, x in zip("ACGT", fields[1:5])},
                    )
        else:
            self.context_model = None

    @staticmethod
    def _disambiguate(sequence):
        r"""Generator of all possible nt sequences implied by a sequence
        containing ambiguous bases."""
        return _sequence_disambiguations(sequence)

    def mutability(self, kmer: str) -> Tuple[np.float64, np.float64]:
        r"""Returns the mutability of a central base of :math:`k`-mer, along
        with nucleotide bias averages over ambiguous ``"N"`` nucleotide
        identities.

        Args:
            kmer: nucleotide :math:`k`-mer
        """
        if self.context_model is None:
            raise ValueError("kmer mutability only defined for context models")
        if len(kmer) != self.k:
            raise ValueError(
                "kmer of length {} inconsistent with context model kmer length {}".format(
                    len(kmer), self.k
                )
            )
        if not all(n in "ACGTN" for n in kmer):
            raise ValueError(
                "sequence {} must contain only characters A, C, G, T, or N".format(kmer)
            )

        cached = self.context_model.get(kmer, None)
        if cached is None:
            mutabilities_to_average, substitutions_to_average = zip(
                *[self.context_model[x] for x in MutationModel._disambiguate(kmer)]
            )

            average_mutability = np.mean(mutabilities_to_average)
            average_substitution = {
                b: sum(
                    substitution_dict[b]
                    for substitution_dict in substitutions_to_average
                )
                / len(substitutions_to_average)
                for b in "ACGT"
            }
            cached = average_mutability, average_substitution
            self.context_model[kmer] = cached

        return cached

    def mutabilities(self, sequence: str) -> List[Tuple[np.float64, np.float64]]:
        r"""Returns the mutability of a sequence at each site, along with
        nucleotide biases.

        Args:
            sequence: nucleotide sequence
        """
        if self.context_model is None:
            return [
                (1, dict((n2, 1 / 3) if n2 is not n else (n2, 0.0) for n2 in "ACGT"))
                for n in sequence
            ]
        else:
            # pad with Ns to allow averaged edge effects
            sequence = "N" * (self.k // 2) + sequence + "N" * (self.k // 2)
            # mutabilities of each nucleotide
            return [
                self.mutability(sequence[(i - self.k // 2) : (i + self.k // 2 + 1)])
                for i in range(self.k // 2, len(sequence) - self.k // 2)
            ]

    def mutate(
        self, sequence: str, lambda0: np.float64 = 1, frame: Optional[int] = None
    ) -> str:
        r"""Mutate a sequence, with lamdba0 the baseline mutability. Cannot
        mutate the same position multiple times.

        Args:
            sequence: nucleotide sequence to mutate
            lambda0: a baseline mutation rate
            frame: the reading frame of the first postition
        """
        sequence_length = len(sequence)
        if frame is not None:
            codon_start = frame - 1
            codon_end = codon_start + 3 * ((sequence_length - codon_start) // 3)
            if "*" in Seq(sequence[codon_start:codon_end]).translate():
                raise RuntimeError("sequence contains stop codon!")

        mutabilities = self.mutabilities(sequence)
        sequence_mutability = (
            sum(mutability[0] for mutability in mutabilities) / sequence_length
        )
        # poisson rate for this sequence (given its relative mutability)
        lambda_sequence = sequence_mutability * lambda0
        # number of mutations m
        trials = 20
        for trial in range(1, trials + 1):
            m = np.random.poisson(lambda_sequence)
            if m <= sequence_length or self.with_replacement:
                break
            if trial == trials:
                raise RuntimeError("mutations saturating, consider reducing lambda0")

        # mutate the sites with mutations
        # if frame is not None and sequence contains stop codon, try again, up to 10 times
        unmutated_positions = range(sequence_length)
        for i in range(m):
            sequence_list = list(sequence)  # make string a list so we can modify it
            # Determine the position to mutate from the mutability matrix
            mutability_p = np.array(
                [mutabilities[pos][0] for pos in unmutated_positions]
            )
            for trial in range(1, trials + 1):
                mut_pos = np.random.choice(
                    unmutated_positions, p=mutability_p / mutability_p.sum()
                )
                # Now draw the target nucleotide using the substitution matrix
                substitution_p = [mutabilities[mut_pos][1][n] for n in "ACGT"]
                assert 0 <= abs(sum(substitution_p) - 1.0) < 1e-10
                chosen_target = np.random.choice(4, p=substitution_p)
                original_base = sequence_list[mut_pos]
                sequence_list[mut_pos] = "ACGT"[chosen_target]
                sequence = "".join(sequence_list)  # reconstruct our sequence
                if (
                    frame is None
                    or "*" not in Seq(sequence[codon_start:codon_end]).translate()
                ):
                    if self.mutation_order:
                        # if mutation order matters, the mutabilities of the sequence need to be updated
                        mutabilities = self.mutabilities(sequence)
                    if not self.with_replacement:
                        # Remove this position so we don't mutate it again
                        unmutated_positions.remove(mut_pos)
                    break
                if trial == trials:
                    raise RuntimeError(
                        "stop codon in simulated sequence on "
                        + str(trials)
                        + " consecutive attempts"
                    )
                sequence_list[mut_pos] = (
                    original_base  # <-- we only get here if we are retrying
                )

        return sequence

    def simulate(
        self,
        sequence: str,
        seq_bounds: Optional[Tuple[Tuple[int, int], Tuple[int, int]]] = None,
        fitness_function: Callable = lambda seq: 0.9,
        lambda0: List[np.float64] = [1],
        frame: Optional[int] = None,
        N_init: int = 1,
        N: Optional[int] = None,
        T: Optional[int] = None,
        n: Optional[int] = None,
        verbose: bool = False,
    ) -> TreeNode:
        r"""Simulate a neutral binary branching process with the mutation model,
        returning a :class:`ete3.Treenode` object.

        Args:
            sequence: root nucleotide sequence
            seq_bounds: ranges for two subsequences used as two parallel genes
            fitness_function: mean number offspring as a function of sequence
            lambda0: baseline mutation rate(s)
            frame: coding frame of starting position(s)
            N_init: initial naive abundnace
            N: maximum population size
            T: maximum generation time
            n: sample size
            verbose: print more messages
        """
        # Checking the validity of the input parameters:
        if N is not None and T is not None:
            raise ValueError("Only one of N and T can be used. One must be None.")
        elif N is None and T is None:
            raise ValueError("Either N or T must be specified.")
        if N is not None and n is not None and n > N:
            raise ValueError("n ({}) must not larger than N ({})".format(n, N))

        # Planting the tree:
        tree = TreeNode()
        tree.dist = 0
        tree.add_feature("sequence", sequence)
        tree.add_feature("terminated", False)
        tree.add_feature("abundance", 0)
        tree.add_feature("time", 0)
        # add fitness attribute, interpreted as mean of offspring distribution
        tree.add_feature("fitness", fitness_function(tree.sequence))

        if N_init > 1:
            for _ in range(N_init):
                child = TreeNode()
                child.dist = 0
                child.add_feature("sequence", sequence)
                child.add_feature("abundance", 0)
                child.add_feature("terminated", False)
                child.add_feature("time", 0)
                # add fitness attribute, interpreted as mean of offspring distribution
                child.add_feature("fitness", fitness_function(child.sequence))
                tree.add_child(child)

        t = 0  # <-- time
        leaves_unterminated = N_init
        while (
            leaves_unterminated > 0
            and (leaves_unterminated < N if N is not None else True)
            and (t < max(T) if T is not None else True)
        ):
            if verbose:
                print("At time:", t)
            t += 1
            list_of_leaves = list(tree.iter_leaves())
            random.shuffle(list_of_leaves)
            for leaf in list_of_leaves:
                # add fitness attribute, interpreted as mean of offspring distribution
                leaf.add_feature("fitness", fitness_function(leaf.sequence))
                if not leaf.terminated:
                    n_children = poisson(leaf.fitness).rvs()
                    leaves_unterminated += (
                        n_children - 1
                    )  # <-- this kills the parent if we drew a zero
                    if not n_children:
                        leaf.terminated = True
                    for child_count in range(n_children):
                        # If sequence pair mutate them separately with their own mutation rate:
                        if seq_bounds is not None:
                            mutated_sequence1 = self.mutate(
                                leaf.sequence[seq_bounds[0][0] : seq_bounds[0][1]],
                                lambda0=lambda0[0],
                                frame=frame,
                            )
                            mutated_sequence2 = self.mutate(
                                leaf.sequence[seq_bounds[1][0] : seq_bounds[1][1]],
                                lambda0=lambda0[1],
                                frame=frame,
                            )
                            mutated_sequence = mutated_sequence1 + mutated_sequence2
                        else:
                            mutated_sequence = self.mutate(
                                leaf.sequence, lambda0=lambda0[0], frame=frame
                            )
                        child = TreeNode()
                        child.dist = utils.hamming_distance(
                            mutated_sequence, leaf.sequence
                        )
                        child.add_feature("sequence", mutated_sequence)
                        child.add_feature("abundance", 0)
                        child.add_feature("terminated", False)
                        child.add_feature("time", t)
                        leaf.add_child(child)

        if N is not None and leaves_unterminated < N:
            raise RuntimeError(
                "tree terminated with {} leaves, {} desired".format(
                    leaves_unterminated, N
                )
            )

        # each leaf in final generation gets an observed abundance of 1, unless downsampled
        if T is not None and len(T) > 1:
            # Iterate the intermediate time steps:
            for Ti in sorted(T)[:-1]:
                # Only sample those that have been 'sampled' at intermediate sampling times:
                final_leaves = [
                    leaf
                    for leaf in tree.iter_descendants()
                    if leaf.time == Ti and leaf.sampled
                ]
                if len(final_leaves) < n:
                    raise RuntimeError(
                        "tree terminated with {} leaves, less than what desired after downsampling {}".format(
                            leaves_unterminated, n
                        )
                    )
                for (
                    leaf
                ) in (
                    final_leaves
                ):  # No need to down-sample, this was already done in the simulation loop
                    leaf.abundance = 1
        # Do the normal sampling of the last time step:
        final_leaves = [leaf for leaf in tree.iter_leaves() if leaf.time == t]
        # by default, downsample to the target simulation size
        if n is not None and len(final_leaves) >= n:
            for leaf in random.sample(final_leaves, n):
                leaf.abundance = 1
        elif n is None and N is not None:
            for leaf in random.sample(final_leaves, N):
                leaf.abundance = 1
        elif N is None and T is not None:
            for leaf in final_leaves:
                leaf.abundance = 1
        elif n is not None and len(final_leaves) < n:
            raise RuntimeError(
                "tree terminated with {} leaves, less than what desired after downsampling {}".format(
                    leaves_unterminated, n
                )
            )
        else:
            raise RuntimeError("Unknown option.")

        # prune away lineages that are unobserved
        for node in tree.iter_descendants():
            if sum(node2.abundance for node2 in node.traverse()) == 0:
                node.detach()

        # # remove unobserved unifurcations
        # for node in tree.iter_descendants():
        #     parent = node.up
        #     if node.abundance == 0 and len(node.children) == 1:
        #         node.delete(prevent_nondicotomic=False)
        #         node.children[0].dist = hamming_distance(node.children[0].sequence, parent.sequence)

        # assign unique names to each node
        for i, node in enumerate(tree.traverse(), 1):
            node.name = "simcell_{}".format(i)

        # return the uncollapsed tree
        return tree


def _sequence_disambiguations(sequence, _accum=""):
    """Iterates through possible disambiguations of sequence, recursively.

    Recursion-depth-limited by number of ambiguity codes in sequence,
    not sequence length.
    """
    if sequence:
        for index, base in enumerate(sequence):
            if base in utils.bases:
                _accum += base
            else:
                for newbase in utils.ambiguous_dna_values[base]:
                    yield from _sequence_disambiguations(
                        sequence[index + 1 :], _accum=(_accum + newbase)
                    )
                return
    yield _accum


def _mutability_dagfuncs(
    *args, splits: List[int] = [], **kwargs
) -> hdag.utils.HistoryDagFilter:
    """Return functions for counting mutability parsimony on the history DAG.

    Mutability parsimony of a tree is the sum over all edges in the tree
    of mutability distances between parent and child node sequences.

    The mutability distance from an ancestral sequence to a target sequence is the sum of
    :math:`-log(mutability * p)` over all sites which do match, where :math:`mutability`
    is the mutation frequency of the k-mer surrounding the mutated base
    (in the ancestral sequence), and :math:`p` is the transition probability to the new
    base in the target sequence. Notice that this so-called distance is not symmetric.

    These functions may not be stable numerically, but we expect every tree to have
    a unique mutability parsimony score for non-degenerate mutability models, so
    so this shouldn't matter in practice.

    Args:
        splits: A list of indices at which sequences are concatenated, but are not
            adjacent in the genome.
        Other arguments are passed to :meth:`mutation_model.MutationModel` constructor.

    Returns:
        A :meth:`historydag.utils.AddFuncDict` which may be passed as keyword arguments
        to :meth:`historydag.HistoryDag.weight_count`, :meth:`historydag.HistoryDag.trim_optimal_weight`,
        or :meth:`historydag.HistoryDag.optimal_weight_annotate`
        methods to trim or annotate a :meth:`historydag.HistoryDag` according to mutability model parsimony.
        Weight format is ``float``.
    """

    mutation_model = MutationModel(*args, **kwargs)
    dist = _mutability_distance(mutation_model, splits=splits)

    def distance(node1, node2):
        if node1.is_root():
            return 0
        else:
            return dist(node1.label.sequence, node2.label.sequence)

    return hdag.utils.HistoryDagFilter(
        hdag.utils.AddFuncDict(
            {
                "start_func": lambda n: 0,
                "edge_weight_func": distance,
                "accum_func": sum,
            },
            name="Mut. Pars.",
        ),
        min,
    )


def _mutability_distance_precursors(
    mutation_model: MutationModel, splits: List[int] = []
):
    chunk_idxs = list(zip([0] + splits, splits + [None]))

    h = mutation_model.k // 2

    # Pads sequence with N's, including in the chain-split boundary to
    # avoid unrelated sites from being treated as part of each others' context.

    # Indices at which padding N's will be in sequences returned from add_ns.
    # Does not include indices of last two N's.
    padding_indices = set(
        itertools.chain.from_iterable(
            [
                range(split + idx * h, split + (idx + 1) * h)
                for idx, split in enumerate([0] + splits)
            ]
        )
    )

    def add_ns(seq: str):
        ns = "N" * h
        chunks = [seq[start:end] for start, end in chunk_idxs]
        return ns + ns.join(chunks) + ns

    @utils._check_distance_arguments
    def mutpairs(seq1: str, seq2: str):
        seq1N = add_ns(seq1)
        seq2N = add_ns(seq2)
        mut_idxs = [
            index
            for index, (base1, base2) in enumerate(zip(seq1N, seq2N))
            if base1 != base2
        ]
        return FrozenMultiset((seq1N[i - h : i + h + 1], seq2N[i]) for i in mut_idxs)

    def sum_minus_logp(pairs: FrozenMultiset):
        # I have chosen to multiply substitution rate for central base
        # with rate of new base. Not sure if this is a good choice.
        if pairs:
            # for floating point behavior
            pairs = sorted(pairs.items())
            p_arr = [
                mult
                * (
                    np.log(mutation_model.mutability(mer)[0])
                    + np.log(mutation_model.mutability(mer)[1][newbase])
                )
                for (mer, newbase), mult in pairs
            ]
            return -sum(p_arr)
        else:
            return 0.0

    def mutability_sum(parent_seq):
        padded_seq = add_ns(parent_seq)
        for idx in padding_indices:
            assert padded_seq[idx] == "N"
        return sum(
            mutation_model.mutability(padded_seq[idx - h : idx + h + 1])[0]
            for idx, _ in enumerate(padded_seq[:-h])
            if idx not in padding_indices
        )

    return (mutpairs, sum_minus_logp, mutability_sum)


def _mutability_distance(mutation_model: MutationModel, splits=[]):
    """Returns a fast distance function based on passed mutation_model.

    First, caches computed mutabilities for k-mers with k // 2 N's on either end. This
    is pretty fast for k=5, but the distance function should be created once
    and reused.

    The returned distance function sums :math:`-log(mutability * p)` over all
    sites which do match between its two sequence arguments, where ``mutability``
    is the mutation frequency of the k-mer surrounding the mutated base (in the
    first sequence argument) and ``p`` is the transition probability to the new
    base.

    Note that, in particular, this function is not symmetric on its  arguments.
    """
    mutpairs, sum_minus_logp, _ = _mutability_distance_precursors(
        mutation_model, splits=splits
    )

    def distance(seq1, seq2):
        return sum_minus_logp(mutpairs(seq1, seq2))

    return distance


def _context_poisson_likelihood(mutation_model: MutationModel, splits=[]):
    mutpairs, sum_minus_logp, mutability_sum = _mutability_distance_precursors(
        mutation_model, splits=splits
    )

    def distance(seq1, seq2):
        subs = mutpairs(seq1, seq2)
        sub_count = len(subs)
        if sub_count == 0:
            return 0
        else:
            mut_sum = mutability_sum(seq1)
            substitution_sum = -sum_minus_logp(subs)
            return (
                substitution_sum
                + (sub_count * (math.log(sub_count) - math.log(mut_sum)))
                - sub_count
            )

    return distance


def _context_poisson_likelihood_dagfuncs(*args, splits: List[int] = [], **kwargs):
    mutation_model = MutationModel(*args, **kwargs)
    distance = _context_poisson_likelihood(mutation_model, splits=splits)

    return hdag.utils.HistoryDagFilter(
        hdag.utils.AddFuncDict(
            {
                "start_func": lambda n: 0,
                "edge_weight_func": lambda n1, n2: (
                    0
                    if n1.is_ua_node()
                    else distance(n1.label.sequence, n2.label.sequence)
                ),
                "accum_func": sum,
            },
            name="LogContextLikelihood",
        ),
        max,
    )



import ete3
import gctree
import numpy as np
import pytest

tree = ete3.TreeNode(name="naive", dist=0)
tree.abundance = 0
tree.sequence = "A"
tree.isotype = {}

child1 = ete3.TreeNode(name="seq1", dist=1)
child1.abundance = 1
child1.sequence = "C"
child1.isotype = {}

child2 = ete3.TreeNode(name="seq2", dist=2)
child2.abundance = 2
child2.sequence = "G"
child2.isotype = {}

grandchild = ete3.TreeNode(name="seq3", dist=3)
grandchild.abundance = 3
grandchild.sequence = "T"
grandchild.isotype = {}

tree.add_child(child1)
tree.add_child(child2)
child1.add_child(grandchild)

ctree = gctree.CollapsedTree()
ctree.tree = tree

τ = τ0 = 1

ctree.local_branching(tau=τ, tau0=τ0)

# dummy integrated branch length for exploded clonal abundances
clone_contribution = τ * (1 - np.exp(-τ0 / τ))

LB_down = {
    tree: {
        tree: 0,
        child1: τ * (1 - np.exp(-(child1.dist + grandchild.dist) / τ))
        + np.exp(-child1.dist / τ) * child1.abundance * clone_contribution
        + np.exp(-(child1.dist + grandchild.dist) / τ)
        * grandchild.abundance
        * clone_contribution,
        child2: τ * (1 - np.exp(-child2.dist / τ))
        + np.exp(-child2.dist / τ) * child2.abundance * clone_contribution,
    },
    child1: {
        child1: child1.abundance * clone_contribution,
        grandchild: τ * (1 - np.exp(-grandchild.dist / τ))
        + np.exp(-grandchild.dist / τ) * grandchild.abundance * clone_contribution,
    },
    child2: {child2: child2.abundance * clone_contribution},
    grandchild: {grandchild: grandchild.abundance * clone_contribution},
}

LB_up = {tree: τ}
LB_up[child1] = τ * (1 - np.exp(-child1.dist / τ)) + np.exp(-child1.dist / τ) * (
    LB_up[tree]
    + sum(LB_down[tree][message] for message in LB_down[tree] if message != child1)
)
LB_up[child2] = τ * (1 - np.exp(-child2.dist / τ)) + np.exp(-child2.dist / τ) * (
    LB_up[tree]
    + sum(LB_down[tree][message] for message in LB_down[tree] if message != child2)
)
LB_up[grandchild] = τ * (1 - np.exp(-grandchild.dist / τ)) + np.exp(
    -grandchild.dist / τ
) * (
    LB_up[child1]
    + sum(
        LB_down[child1][message] for message in LB_down[child1] if message != grandchild
    )
)

LBI = {node: LB_up[node] + sum(LB_down[node].values()) for node in tree.traverse()}
LBR = {node: sum(LB_down[node].values()) / LB_up[node] for node in tree.traverse()}

print("name\tLBI\tLBR")
for node in ctree.tree.traverse():
    print(f"{node.name}\t{node.LBI:.3f}\t{node.LBR:.3f}")
    assert LB_up[node] == pytest.approx(node.LB_up)
    assert LB_down[node] == pytest.approx(node.LB_down)
    assert LBI[node] == pytest.approx(node.LBI)
    assert LBR[node] == pytest.approx(node.LBR) or (
        np.isnan(LBR[node]) and np.isnan(node.LBR)
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""Read a PHYLIP-format file and produce an appropriate config file for passing
to `dnapars`.

`dnapars` doesn't play very well in a pipeline.  It prompts the user for
configuration information and reads responses from stdin.  The config
file generated by this script is meant to mimic the responses to the
expected prompts.

Typical usage is,

$ mkconfig sequence.phy > dnapars.cfg $ dnapars < dnapars.cfg
"""
import os
import argparse


def get_parser():
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument("phylip", help="PHYLIP input", type=str)
    parser.add_argument("treeprog", help="dnapars, dnaml, or seqboot", type=str)
    parser.add_argument(
        "--quick", action="store_true", help="quicker (less thourough) dnapars"
    )
    parser.add_argument(
        "--jumble",
        default=10,
        help="search tree space with this many random permutations of the input sequences",
    )
    parser.add_argument(
        "--bootstrap",
        type=int,
        default=0,
        help="input is seqboot output with this many samples",
    )
    return parser


def main(arg_list=None):
    args = get_parser().parse_args(arg_list)
    print(os.path.realpath(args.phylip))  # phylip input file
    if args.treeprog == "seqboot":
        print("R")
        print(args.bootstrap)
        print("Y")
        # random seed for bootstrap (odd integer)
        print("1")
        return
    print("J")
    # random seed for tree search
    print("1")
    print(args.jumble)
    if args.bootstrap:
        print("M")
        print("D")
        print(args.bootstrap)
    if args.treeprog == "dnapars":
        print("O")  # Outgroup root
        print(1)  # arbitrary root on first
        if args.quick:
            print("S")
            print("Y")
        print("4")
        print("5")
        print(".")
        print("Y")
    elif args.treeprog == "dnaml":
        print("O")  # Outgroup root
        print(1)  # arbitrary root on first
        print("R")  # gamma
        print("5")  # Reconstruct hypothetical seq
        print("Y")  # accept these
        print("1.41421356237")  # CV = sqrt(2) (alpha = .5)
        print("4")  # 4 catagories
    else:
        raise RuntimeError(
            "treeprog=" + args.treeprog + ' is not "dnaml", "dnapars", or "seqboot"'
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Given an outputfile from one of the PHYLIP tools - `dnaml` or `dnapars` -
produce a CollapsedForest containing the trees in that outputfile."""

import gctree.branching_processes as bp
import gctree.utils

from ete3 import Tree
import re
import random
from collections import defaultdict

import pickle
import argparse
from typing import Dict

code_vectors = {
    code: [
        0 if base in gctree.utils.ambiguous_dna_values[code] else float("inf")
        for base in gctree.utils.bases
    ]
    for code in gctree.utils.ambiguous_dna_values
}
cost_adjust = {
    base: [int(not i == j) for j in range(5)]
    for i, base in enumerate(gctree.utils.bases)
}


# iterate over recognized sections in the phylip output file.
def sections(fh):
    patterns = {
        "parents": r"\s+between\s+and\s+length",
        ("sequences", "dnaml"): r"\s*node\s+reconstructed\s+sequence",
        ("sequences", "dnapars"): r"from\s+to\s+any steps",
        "seqboot_dataset": r"Data\s*set",
    }
    patterns = {k: re.compile(v, re.IGNORECASE) for (k, v) in patterns.items()}
    for line in fh:
        for k, pat in patterns.items():
            if pat.match(line):
                yield k
                break


# iterate over entries in the distance section
def iter_edges(fh):
    #  152          >root2           0.01208     (     zero,     0.02525) **
    pat = re.compile(
        r"\s*(?P<parent>\w+)\s+(?P<child>[\w>_.-]+)\s+(?P<distance>\d+\.\d+)"
    )
    # drop the header underline
    fh.readline()
    matches = 0
    for line in fh:
        m = pat.match(line)
        if m:
            matches += 1
            yield (m.group("child"), m.group("parent"))
        # We only want to break at the very end of the block of matches; dnaml has an extra blank line between
        # header and rows that dnapars doesn't
        elif not matches:
            continue
        else:
            break


# iterate over entries in the sequences section
def parse_seqdict(fh, mode="dnaml"):
    #  152        sssssssssG AGGTGCAGCT GTTGGAGTCT GGGGGAGGCT TGGTACAGCC TGGGGGGTCC
    seqs = defaultdict(str)
    if mode == "dnaml":
        patterns = re.compile(r"^\s*(?P<id>[a-zA-Z0-9>_.-]*)\s+(?P<seq>[a-zA-Z \-]+)")
    elif mode == "dnapars":
        patterns = re.compile(
            r"^\s*\S+\s+(?P<id>[a-zA-Z0-9>_.-]*)\s+(yes\s+|no\s+|maybe\s+)?(?P<seq>[a-zA-Z \-\?]+)"
        )
    else:
        raise ValueError("invalid mode " + mode)
    fh.readline()
    for line in fh:
        m = patterns.match(line)
        if m and m.group("id") != "":
            last_blank = False
            seqs[m.group("id")] += m.group("seq").replace(" ", "").upper()
        elif line.rstrip() == "":
            if last_blank:
                break
            else:
                last_blank = True
                continue
        else:
            break
    return seqs


# parse the dnaml output file and return data structures containing a
# list biopython.SeqRecords and a dict containing adjacency
# relationships and distances between nodes.
def parse_outfile(outfile, abundance_file=None, root="root", disambiguate=False):
    """Parse phylip outfile, and return dnapars trees."""
    if abundance_file is not None:
        counts = {
            line.split(",")[0]: int(line.split(",")[1]) for line in open(abundance_file)
        }
    # No count, just make an empty count dictionary:
    else:
        counts = None
    trees = []
    bootstrap = False
    # Ugg... for compilation need to let python know that these will definely both be defined :-/
    sequences, parents = {}, {}
    with open(outfile, "r") as fh:
        for sect in sections(fh):
            if sect == "parents":
                parents = {child: parent for child, parent in iter_edges(fh)}
            elif sect[0] == "sequences":
                sequences = parse_seqdict(fh, sect[1])
                # sanity check;  a valid tree should have exactly one node that is parentless
                if not len(parents) == len(sequences) - 1:
                    raise RuntimeError(
                        "invalid results attempting to parse {}: there are {} parentless sequences".format(
                            outfile, len(sequences) - len(parents)
                        )
                    )
                if bootstrap:
                    trees[-1].append(build_tree(sequences, parents, counts, root))
                else:
                    trees.append(build_tree(sequences, parents, counts, root))
            elif sect == "seqboot_dataset":
                bootstrap = True
                trees.append([])
            else:
                raise RuntimeError("unrecognized phylip section = {}".format(sect))
    if len(trees) == 0:
        raise RuntimeError(f"No trees found in '{outfile}'")
    if disambiguate:
        # Disambiguate sets node.dist for all nodes in disambiguated trees
        trees = [disambiguate(tree) for tree in trees]
    return trees


def disambiguate(tree: Tree, random_state=None) -> Tree:
    """Randomly resolve ambiguous bases using a two-pass Sankoff Algorithm on
    subtrees of consecutive ambiguity codes."""
    if random_state is None:
        random.seed(tree.write(format=1))
    else:
        random.setstate(random_state)
    for node in tree.traverse():
        for site, base in enumerate(node.sequence):
            if base not in gctree.utils.bases:

                def is_leaf(node):
                    return (node.is_leaf()) or (
                        node.sequence[site] in gctree.utils.bases
                    )

                # First pass of Sankoff: compute cost vectors
                for node2 in node.traverse(strategy="postorder", is_leaf_fn=is_leaf):
                    base2 = node2.sequence[site]
                    node2.add_feature("cv", code_vectors[base2].copy())
                    if not is_leaf(node2):
                        for i in range(5):
                            for child in node2.children:
                                node2.cv[i] += min(
                                    [
                                        sum(v)
                                        for v in zip(
                                            child.cv, cost_adjust[gctree.utils.bases[i]]
                                        )
                                    ]
                                )
                # Second pass: Choose base and adjust children's cost vectors
                if not node.is_root():
                    node.cv = [
                        sum(v)
                        for v in zip(node.cv, cost_adjust[node.up.sequence[site]])
                    ]
                # traverse evaluates is_leaf(node) after yielding node.
                # Resolving base makes is_leaf true; must get order before
                # making changes.
                preorder = list(node.traverse(strategy="preorder", is_leaf_fn=is_leaf))
                for node2 in preorder:
                    if node2.sequence[site] in gctree.utils.bases:
                        continue
                    min_cost = min(node2.cv)
                    base_index = random.choice(
                        [i for i, val in enumerate(node2.cv) if val == min_cost]
                    )
                    new_base = gctree.utils.bases[base_index]
                    # Adjust child cost vectors
                    if not is_leaf(node2):
                        for child in node2.children:
                            child.cv = [
                                sum(v) for v in zip(child.cv, cost_adjust[new_base])
                            ]
                    node2.sequence = (
                        node2.sequence[:site] + new_base + node2.sequence[(site + 1) :]
                    )
    for node in tree.traverse():
        try:
            node.del_feature("cv")
        except (AttributeError, KeyError):
            pass
    tree.dist = 0
    for node in tree.iter_descendants():
        node.dist = gctree.utils.hamming_distance(node.up.sequence, node.sequence)
    return tree


# build a tree from a set of sequences and an adjacency dict.
def build_tree(
    sequences: Dict[str, str], parents: Dict[str, str], counts=None, root="root"
):
    """Build an ete tree from sequences and parents dictionaries.

    Args:
        sequences: a dictionary mapping names to sequences
        parents: a dictionary mapping parent node names to child node names
        counts: a dictionary mapping node names to observed abundances. This argument
            is no longer used in the main gctree inference pipeline (counts are assigned in DAG)
            but remains for compatibility.
        root: the name of the root node
    """
    # build an ete tree
    # first a dictionary of disconnected nodes
    nodes = {}
    for name in sequences:
        node = Tree()
        node.name = name
        node.add_feature("sequence", sequences[node.name])
        if counts is not None:
            if node.name in counts:
                node.add_feature("abundance", counts[node.name])
            else:
                node.add_feature("abundance", 0)
        nodes[name] = node
    for name in sequences:
        if name in parents:
            nodes[parents[name]].add_child(nodes[name])
        else:
            tree = nodes[name]
    # reroot on root
    if root is not None:
        try:
            root_id = [node for node in nodes if root in node][0]
        except IndexError:
            raise RuntimeError(f"Provided root id '{root}' not found in dnapars tree.")
        assert len(nodes[root_id].children) == 0
        root_parent = nodes[root_id].up
        root_parent.remove_child(nodes[root_id])
        nodes[root_id].add_child(root_parent)
        # remove possible unecessary unifurcation after rerooting
        if len(root_parent.children) == 1:
            root_parent.delete(prevent_nondicotomic=False)
            root_parent.children[0].dist = gctree.utils.hamming_distance(
                root_parent.children[0].sequence, nodes[root_id].sequence
            )
        tree = nodes[root_id]

    return tree


def get_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "phylip_outfile",
        type=str,
        help="dnaml outfile (verbose output with inferred ancestral sequences, option 5).",
    )
    parser.add_argument("abundance_file", type=str, help="count file")
    parser.add_argument("--outputfile", default=None, help="output file.")
    parser.add_argument("--root", default="root", help="root sequence id")
    return parser


def main(arg_list=None):
    args = get_parser().parse_args(arg_list)

    if args.outputfile is None:
        args.outputfile = args.phylip_outfile + ".collapsed_forest.p"
    forest = bp.CollapsedForest(
        parse_outfile(args.phylip_outfile, args.abundance_file, args.root)
    )
    with open(args.outputfile, "wb") as fh:
        pickle.dump(forest, fh)


if __name__ == "__main__":
    main()



import random
import ete3
import gctree.phylip_parse as pps

newick_tree1 = (
    "((((12[&&NHX:name=12:sequence=T])4[&&NHX:name=4:sequence=C],"
    "(6[&&NHX:name=6:sequence=C],"
    "7[&&NHX:name=7:sequence=A])5[&&NHX:name=5:sequence=M])"
    "3[&&NHX:name=3:sequence=M],8[&&NHX:name=8:sequence=A],"
    "(11[&&NHX:name=11:sequence=A],10[&&NHX:name=10:sequence=G])"
    "9[&&NHX:name=9:sequence=R])2[&&NHX:name=2:sequence=R])"
    "1[&&NHX:name=1:sequence=G];"
)

newick_tree2 = (
    "((((12[&&NHX:name=12:sequence=T])4[&&NHX:name=4:sequence=C],"
    "(6[&&NHX:name=6:sequence=C],"
    "7[&&NHX:name=7:sequence=A])5[&&NHX:name=5:sequence=?])"
    "3[&&NHX:name=3:sequence=?],8[&&NHX:name=8:sequence=A],"
    "(11[&&NHX:name=11:sequence=A],10[&&NHX:name=10:sequence=G])"
    "9[&&NHX:name=9:sequence=?])2[&&NHX:name=2:sequence=?])"
    "1[&&NHX:name=1:sequence=G];"
)


def treeprint(tree: ete3.TreeNode):
    tree = tree.copy()
    for node in tree.traverse():
        node.name = node.sequence
    return tree.write(format=8)


def sample(treestring, n=20):
    ogt = ete3.TreeNode(newick=newick_tree2, format=1)
    print(ogt.sequence)
    newickset = set()
    for i in range(n):
        t = ogt.copy()
        random.seed(i)
        t = pps.disambiguate(t, random_state=random.getstate())
        newickset.add(treeprint(t))
    return newickset


def test_full_ambiguity():
    newickset = sample(newick_tree2)
    correctset = {
        "((((T)C,(C,A)C)C,A,(A,G)G)G);",
        "((((T)C,(C,A)A)A,A,(A,G)A)A);",
        "((((T)C,(C,A)C)C,A,(A,G)A)A);",
    }
    if not newickset == correctset:
        missing = correctset - newickset
        wrong = newickset - correctset
        print(
            f"\nDisambiguate function is missing {missing}\n"
            f"and came up with these incorrect trees: {wrong}"
        )
        raise ValueError("Invalid Disambiguation")


def test_restricted_ambiguity():
    newickset = sample(newick_tree1)
    correctset = {
        "((((T)C,(C,A)C)C,A,(A,G)G)G);",
        "((((T)C,(C,A)A)A,A,(A,G)A)A);",
        "((((T)C,(C,A)C)C,A,(A,G)A)A);",
    }
    if not newickset == correctset:
        missing = correctset - newickset
        wrong = newickset - correctset
        print(
            f"\nDisambiguate function is missing {missing}\n"
            f"and came up with these incorrect trees: {wrong}"
        )
        raise ValueError("Invalid Disambiguation")



r"""This module contains classes for simulation and inference for a binary
branching process with mutation in which the tree is collapsed to nodes that
count the number of clonal leaves of each type."""

from __future__ import annotations

import gctree.utils
from gctree.isotyping import _isotype_dagfuncs, _isotype_annotation_dagfuncs
from gctree.mutation_model import (
    _mutability_dagfuncs,
    _context_poisson_likelihood_dagfuncs,
)
from gctree.phylip_parse import disambiguate

from frozendict import frozendict
import pandas as pd
import seaborn as sns
import numpy as np
import warnings
import random
import os
import scipy.special as scs
import scipy.optimize as sco
import ete3
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import AlignIO
from Bio.Phylo.TreeConstruction import MultipleSeqAlignment
import pickle
import functools
import collections as coll
import historydag as hdag
import multiset
import matplotlib as mp
import matplotlib.pyplot as plt
from typing import Tuple, Dict, List, Union, Set, Callable, Mapping, Sequence, Optional
from decimal import Decimal
import math

sequence_resolutions = hdag.parsimony_utils.standard_nt_ambiguity_map_gap_as_char.get_sequence_resolution_func(
    "sequence"
)
sequence_resolutions_count = hdag.parsimony_utils.standard_nt_ambiguity_map_gap_as_char.get_sequence_resolution_count_func(
    "sequence"
)


class CollapsedTree:
    r"""A collapsed tree, modeled as an infinite type Galton-Watson process run
    to extinction.

    Attributes:
        tree: :class:`ete3.TreeNode` object with ``abundance`` node features

    Args:
        tree: ete3 tree with ``abundance`` node features. If uncollapsed,
            it will be collapsed along branches with no mutations.
            Can be ommitted on initializaion, and later simulated.
            If a tree is provided, names of nodes with abundance 0 will not be preserved.
        allow_repeats: tolerate the existence of nodes with the same genotype after collapse, e.g. in sister clades.
    """

    _max_ll_cache: Dict[Tuple[float, float], Tuple[int, int]] = {}

    def __init__(
        self, tree: Optional[ete3.TreeNode] = None, allow_repeats: bool = False
    ):
        if tree is not None:
            self.tree = tree.copy()
            self.tree.dist = 0

            def merge_isotype_dicts(parent_isotype, child_isotype):
                # values are abundances and keys are isotypes.
                parent_isotype = dict(parent_isotype)
                for key, val in child_isotype.items():
                    if key in parent_isotype:
                        parent_isotype[key] = max(parent_isotype[key], val)
                    else:
                        parent_isotype[key] = val
                return frozendict(parent_isotype)

            # remove unobserved internal unifurcations
            for node in self.tree.iter_descendants():
                if node.abundance == 0 and len(node.children) == 1:
                    node.delete(prevent_nondicotomic=False)

            # ensure distances are correct before collapse
            for node in self.tree.iter_descendants():
                node.dist = gctree.utils.hamming_distance(
                    node.sequence, node.up.sequence
                )

            # iterate over the tree below root and collapse edges of zero
            # length if the node is a leaf and it's parent has nonzero
            # abundance we combine taxa names to a set to acommodate
            # bootstrap samples that result in repeated genotypes
            observed_genotypes = set(
                node.name for node in self.tree.traverse() if node.abundance
            )
            observed_genotypes.add(self.tree.name)
            for node in self.tree.get_descendants(strategy="postorder"):
                if node.dist == 0:
                    # if an abundance is nonzero, that's the right one.
                    node.up.abundance = max(node.abundance, node.up.abundance)
                    # isotype is dictionary with isotype as key and observed
                    # abundance as value
                    if "isotype" in node.features:
                        node.up.isotype = merge_isotype_dicts(
                            node.up.isotype, node.isotype
                        )
                    # original_ids is a set of observed ids corresponding to
                    # each node
                    if "original_ids" in node.features:
                        node.up.original_ids = node.original_ids | node.up.original_ids
                    if isinstance(node.name, str):
                        node_set = set([node.name])
                    else:
                        node_set = set(node.name)
                    if isinstance(node.up.name, str):
                        node_up_set = set([node.up.name])
                    else:
                        node_up_set = set(node.up.name)
                    if node_up_set < observed_genotypes:
                        if node_set < observed_genotypes:
                            node.up.name = tuple(node_set | node_up_set)
                            if len(node.up.name) == 1:
                                node.up.name = node.up.name[0]
                    elif node_set < observed_genotypes:
                        node.up.name = tuple(node_set)
                        if len(node.up.name) == 1:
                            node.up.name = node.up.name[0]
                    node.delete(prevent_nondicotomic=False)

                if "isotype" in node.features:
                    node.add_feature(
                        "inferred_isotype", min(node.isotype.keys(), default=None)
                    )
            if "isotype" in node.features:
                self.tree.add_feature(
                    "inferred_isotype", min(self.tree.isotype.keys(), default=None)
                )

            final_observed_genotypes = set()
            for node in self.tree.traverse():
                if node.abundance > 0 or node == self.tree:
                    final_observed_genotypes.add(node.name)
            if final_observed_genotypes != observed_genotypes:
                raise RuntimeError(
                    "observed genotypes don't match after "
                    f"collapse\n\tbefore: {observed_genotypes}"
                    f"\n\tafter: {final_observed_genotypes}\n\t"
                    "symmetric diff: "
                    f"{observed_genotypes ^ final_observed_genotypes}"
                )

            rep_seq = sum(node.abundance > 0 for node in self.tree.traverse()) - len(
                set(
                    [
                        node.sequence
                        for node in self.tree.traverse()
                        if node.abundance > 0
                    ]
                )
            )
            if not allow_repeats and rep_seq:
                raise RuntimeError(
                    "Repeated observed sequences in collapsed "
                    f"tree. {rep_seq} sequences were found repeated."
                )
            elif allow_repeats and rep_seq:
                rep_seq = sum(
                    node.abundance > 0 for node in self.tree.traverse()
                ) - len(
                    set(
                        [
                            node.sequence
                            for node in self.tree.traverse()
                            if node.abundance > 0
                        ]
                    )
                )
                print(
                    "Repeated observed sequences in collapsed tree. "
                    f"{rep_seq} sequences were found repeated."
                )
            # a custom ladderize accounting for abundance and sequence to break
            # ties in abundance. In same traversal fix unobserved node names
            unobserved_count = 1
            unobserved_dict = {}
            for node in self.tree.traverse(strategy="postorder"):
                # add a partition feature and compute it recursively up tree
                node.add_feature(
                    "partition",
                    node.abundance + sum(node2.partition for node2 in node.children),
                )
                # sort children of this node based on partion and sequence
                node.children.sort(key=lambda node: (node.partition, node.sequence))
                # change node name if necessary
                if node.abundance == 0 and not node.is_root():
                    if node.sequence in unobserved_dict:
                        node.name = unobserved_dict[node.sequence]
                    else:
                        node.name = str(unobserved_count)
                        unobserved_count += 1
                        unobserved_dict[node.sequence] = node.name
            self._build_cm_counts()
        else:
            self.tree = None

    def _build_cm_counts(self):
        # create tuple (c, m) for each node, and store in a tuple of
        # ((c, m), n)'s, where n is the multiplicity of (c, m) seen in the
        # tree, adding pseudocount to root if unobserved unifurcation at root.
        cmlist = [
            (node.abundance, len(node.children))
            for node in self.tree.iter_descendants()
        ]
        rootcm = (self.tree.abundance, len(self.tree.children))
        if rootcm == (0, 1):
            cmlist.append((1, 1))
        else:
            cmlist.append(rootcm)
        self._cm_counts = tuple(coll.Counter(cmlist).items())

    @staticmethod
    def _simulate_genotype(p: np.float64, q: np.float64) -> Tuple[int, int]:
        r"""Simulate the number of clonal leaves :math:`c` and mutant clades
        :math:`m` as a Galton-Watson process with branching probability
        :math:`p` and mutation probability :math:`q`.

        Args:
            p: branching probability
            q: mutation probability

        Returns:
            Tuple :math:`(c, m)` of the number of clonal leaves and mutant clades
        """
        if not (0 <= p <= 1 and 0 <= q <= 1):
            raise ValueError(
                "p and q must be in the unit interval: " f"p = {p}, q = {q}"
            )
        if p >= 0.5:
            warnings.warn(
                f"p = {p} is not subcritical, tree simulations not garanteed to terminate!"
            )
        # let's track the tree in breadth first order, listing number of clonal
        # and mutant descendants of each node mutant clades terminate in this
        # view
        cumsum_clones = 0
        len_tree = 0
        c = 0
        m = 0
        # while termination condition not met
        while cumsum_clones > len_tree - 1:
            if random.random() < p:
                mutants = sum(random.random() < q for child in range(2))
                clones = 2 - mutants
                m += mutants
            else:
                mutants = 0
                clones = 0
                c += 1
            cumsum_clones += clones
            len_tree += 1
        assert cumsum_clones == len_tree - 1
        return c, m

    @staticmethod
    @functools.lru_cache(maxsize=None)
    def _ll_genotype(
        c: int, m: int, p: np.float64, q: np.float64
    ) -> Tuple[np.float64, np.ndarray]:
        r"""Log-probability of getting :math:`c` leaves that are clones of the
        root and :math:`m` mutant clades off the root lineage, given branching
        probability :math:`p` and mutation probability :math:`q`.

        AKA the spaceship distribution. Also returns gradient wrt p and q
        (p, q). Computed by dynamic programming.

        Args:
            c: clonal leaves
            m: mutant clades
            p: branching probability
            q: mutation probability

        Returns:
            log-likelihood and gradient wrt :math:`p` and :math:`q`.
        """
        if (p, q) in CollapsedTree._max_ll_cache:
            cached_c, cached_m = CollapsedTree._max_ll_cache[(p, q)]
        else:
            # clear cache for old parameters if parameters change:
            CollapsedTree._max_ll_cache = {}
            CollapsedTree._ll_genotype.cache_clear()
            cached_c, cached_m = 0, 0
        # Check cache is built
        if c > cached_c or m > cached_m:
            CollapsedTree._max_ll_cache[(p, q)] = (c, m)
            # Build missing cache in three parts:
            # |1 3
            # |X 2
            # Where 'X' is already built, axes are c and m.
            for cx in range(cached_c + 1):
                for mx in range(cached_m, m + 1):
                    if cx > 0 or mx > 1:
                        CollapsedTree._ll_genotype(cx, mx, p, q)
            for mx in range(cached_m + 1):
                for cx in range(cached_c, c + 1):
                    if cx > 0 or mx > 1:
                        CollapsedTree._ll_genotype(cx, mx, p, q)
            for mx in range(cached_m + 1, m + 1):
                for cx in range(cached_c + 1, c + 1):
                    if cx > 0 or mx > 1:
                        CollapsedTree._ll_genotype(cx, mx, p, q)
            # If we're here, we've computed what we want
            return CollapsedTree._ll_genotype(c, m, p, q)

        if c == m == 0 or (c == 0 and m == 1):
            raise ValueError("Zero likelihood event")
        elif c == 1 and m == 0:
            logf_result = np.log(1 - p)
            dlogfdp_result = -1 / (1 - p)
            dlogfdq_result = 0
        elif c == 0 and m == 2:
            logf_result = np.log(p) + 2 * np.log(q)
            dlogfdp_result = 1 / p
            dlogfdq_result = 2 / q
        else:
            if m >= 1:
                (
                    neighbor_ll_genotype,
                    (
                        neighbor_dlogfdp,
                        neighbor_dlogfdq,
                    ),
                ) = CollapsedTree._ll_genotype(c, m - 1, p, q)
                logg_array = [
                    (
                        np.log(2)
                        + np.log(p)
                        + np.log(q)
                        + np.log(1 - q)
                        + neighbor_ll_genotype
                    )
                ]
                dloggdp_array = [1 / p + neighbor_dlogfdp]
                dloggdq_array = [1 / q - 1 / (1 - q) + neighbor_dlogfdq]
            else:
                logg_array = []
                dloggdp_array = []
                dloggdq_array = []
            for cx in range(c + 1):
                for mx in range(m + 1):
                    if (cx > 0 or mx > 1) and (c - cx > 0 or m - mx > 1):
                        (
                            neighbor1_ll_genotype,
                            (neighbor1_dlogfdp, neighbor1_dlogfdq),
                        ) = CollapsedTree._ll_genotype(cx, mx, p, q)
                        (
                            neighbor2_ll_genotype,
                            (neighbor2_dlogfdp, neighbor2_dlogfdq),
                        ) = CollapsedTree._ll_genotype(c - cx, m - mx, p, q)
                        logg = (
                            np.log(p)
                            + 2 * np.log(1 - q)
                            + neighbor1_ll_genotype
                            + neighbor2_ll_genotype
                        )
                        dloggdp = 1 / p + neighbor1_dlogfdp + neighbor2_dlogfdp
                        dloggdq = -2 / (1 - q) + neighbor1_dlogfdq + neighbor2_dlogfdq
                        logg_array.append(logg)
                        dloggdp_array.append(dloggdp)
                        dloggdq_array.append(dloggdq)
            if not logg_array:
                raise ValueError("Zero likelihood event")
            else:
                logf_result = scs.logsumexp(logg_array)
                softmax_logg_array = scs.softmax(logg_array)
                dlogfdp_result = np.multiply(softmax_logg_array, dloggdp_array).sum()
                dlogfdq_result = np.multiply(softmax_logg_array, dloggdq_array).sum()

        return (logf_result, np.array([dlogfdp_result, dlogfdq_result]))

    @np.errstate(all="raise")
    def ll(
        self,
        p: np.float64,
        q: np.float64,
    ) -> Tuple[np.float64, np.ndarray]:
        r"""Log likelihood of branching process parameters :math:`(p, q)` given
        tree topology :math:`T` and genotype abundances :math:`A`.

        .. math::
            \ell(p, q; T, A) = \log\mathbb{P}(T, A \mid p, q)

        Args:
            p: branching probability
            q: mutation probability

        Returns:
            Log likelihood :math:`\ell(p, q; T, A)` and its gradient :math:`\nabla\ell(p, q; T, A)`
        """
        if self.tree is None:
            raise ValueError("tree data must be defined to compute likelihood")
        return _lltree(self._cm_counts, p, q)

    def mle(self, **kwargs) -> Tuple[np.float64, np.float64]:
        r"""Maximum likelihood estimate of :math:`(p, q)`.

        .. math::
            (p, q) = \arg\max_{p,q\in [0,1]}\ell(p, q)

        Args:
            kwargs: keyword arguments passed along to the branching process likelihood :meth:`CollapsedTree.ll`

        Returns:
            Tuple :math:`(p, q)` with estimated branching probability and estimated mutation probability
        """
        return _mle_helper(self.ll, **kwargs)

    def simulate(self, p: np.float64, q: np.float64, root: bool = True):
        r"""Simulate a collapsed tree as an infinite type Galton-Watson process
        run to extintion, with branching probability :math:`p` and mutation
        probability :math:`q`. Overwrites existing tree attribute.

        Args:
            p: branching probability
            q: mutation probability
            root: flag indicating simulation is being run from the root of the tree, so we should update tree attributes (should usually be ``True``)
        """
        c, m = self._simulate_genotype(p, q)
        self.tree = ete3.TreeNode()
        self.tree.add_feature("abundance", c)
        for _ in range(m):
            # ooooh, recursion
            child = CollapsedTree()
            child.simulate(p, q, root=False)
            child = child.tree
            child.dist = 1
            self.tree.add_child(child)

        if root:
            # create list of (c, m) for each node
            self._build_cm_counts()

    def __repr__(self):
        r"""Return a string representation for printing."""
        return str(self.tree)

    def render(
        self,
        outfile: str,
        scale: Optional[float] = None,
        branch_margin: float = 0,
        node_size: Optional[float] = None,
        idlabel: bool = False,
        colormap: Optional[Dict] = None,
        frame: Optional[int] = None,
        position_map: Optional[List] = None,
        chain_split: Optional[int] = None,
        frame2: Optional[int] = None,
        position_map2: Optional[List] = None,
        show_support: bool = False,
    ):
        r"""Render to tree image file.

        Args:
            outfile: file name to render to, filetype inferred from suffix, .svg for color
            scale: branch length scale in pixels (set automatically if ``None``)
            branch_margin: additional leaf branch separation margin, in pixels, to scale tree width
            node_size: size of nodes in pixels (set according to abundance if ``None``)
            idlabel: label nodes with seq ids, and write sequences of all nodes to a fasta file with same base name as ``outfile``
            colormap: dictionary mapping node names to color names or to dictionaries of color frequencies
            frame: coding frame for annotating amino acid substitutions
            position_map: mapping of position names for sequence indices, to be used with substitution annotations and the ``frame`` argument
            chain_split: if sequences are a concatenation two gene sequences, this is the index at which the 2nd one starts (requires ``frame`` and ``frame2`` arguments)
            frame2: coding frame for 2nd sequence when using ``chain_split``
            position_map2: like ``position_map``, but for 2nd sequence when using ``chain_split``
            show_support: annotate bootstrap support if available
        """
        if frame is not None and frame not in (1, 2, 3):
            raise RuntimeError("frame must be 1, 2, or 3")

        def my_layout(node):
            if colormap is None or node.name not in colormap:
                circle_color = "lightgray"
            else:
                circle_color = colormap[node.name]
            text_color = "black"

            if node_size is None:
                node_size2 = max(1, 10 * np.sqrt(node.abundance))
                if isinstance(circle_color, str):
                    C = ete3.CircleFace(
                        radius=node_size2,
                        color=circle_color,
                        label=(
                            {"text": str(node.abundance), "color": text_color}
                            if node.abundance > 0
                            else None
                        ),
                    )
                    C.rotation = -90
                    C.hz_align = 1
                    ete3.faces.add_face_to_node(C, node, 0)
                else:
                    P = ete3.PieChartFace(
                        [100 * x / node.abundance for x in circle_color.values()],
                        2 * node_size2,
                        2 * node_size2,
                        colors=[
                            (color if color != "None" else "lightgray")
                            for color in list(circle_color.keys())
                        ],
                        line_color=None,
                    )
                    T = ete3.TextFace(
                        " ".join([str(x) for x in list(circle_color.values())]),
                        tight_text=True,
                    )
                    T.hz_align = 1
                    T.rotation = -90
                    ete3.faces.add_face_to_node(P, node, 0, position="branch-right")
                    ete3.faces.add_face_to_node(T, node, 1, position="branch-right")
            elif node.abundance:
                T = ete3.TextFace(node.abundance)
                T.rotation = -90
                T.hz_align = 1
                ete3.faces.add_face_to_node(
                    T,
                    node,
                    0,
                    position="branch-top",
                )

            if idlabel:
                T = ete3.TextFace(node.name, tight_text=True, fsize=6)
                T.rotation = -90
                T.hz_align = 1
                ete3.faces.add_face_to_node(
                    T,
                    node,
                    1 if isinstance(circle_color, str) else 2,
                    position="branch-bottom",
                )

        # we render on a copy, so faces are not permanent
        tree_copy = self.tree.copy(method="deepcopy")
        for node in tree_copy.traverse():
            nstyle = ete3.NodeStyle()
            if colormap is None or node.name not in colormap:
                nstyle["fgcolor"] = "lightgray"
            else:
                nstyle["fgcolor"] = colormap[node.name]
            if node_size is not None:
                nstyle["size"] = node_size
            else:
                nstyle["size"] = 0
            if node.up is not None:
                if "sequence" in tree_copy.features and set(
                    node.sequence.upper()
                ) == set("ACGT"):
                    if frame is not None:
                        if chain_split is not None and frame2 is None:
                            raise ValueError(
                                "must define frame2 when using chain_split"
                            )
                        if frame2 is not None and chain_split is None:
                            raise ValueError(
                                "must define chain_split when using frame2"
                            )
                        # loop over split heavy/light chain subsequences
                        for start, end, framex, position_mapx in (
                            (0, chain_split, frame, position_map),
                            (chain_split, None, frame2, position_map2),
                        ):
                            if start is not None:
                                seq = node.sequence[start:end]
                                aa = Seq(
                                    seq[
                                        (framex - 1) : (
                                            framex
                                            - 1
                                            + (3 * ((len(seq) - (framex - 1)) // 3))
                                        )
                                    ]
                                ).translate()
                                seq = node.up.sequence[start:end]
                                aa_parent = Seq(
                                    seq[
                                        (framex - 1) : (
                                            framex
                                            - 1
                                            + (3 * ((len(seq) - (framex - 1)) // 3))
                                        )
                                    ]
                                ).translate()
                                mutations = [
                                    f"{aa1}{pos if position_mapx is None else position_mapx[pos]}{aa2}"
                                    for pos, (aa1, aa2) in enumerate(zip(aa_parent, aa))
                                    if aa1 != aa2
                                ]
                                if mutations:
                                    T = ete3.TextFace(
                                        "\n".join(mutations),
                                        fsize=6,
                                        tight_text=False,
                                    )
                                    if start == 0:
                                        T.margin_top = 6
                                    else:
                                        T.margin_bottom = 6
                                    T.rotation = -90
                                    node.add_face(
                                        T,
                                        0,
                                        position=(
                                            "branch-bottom"
                                            if start == 0
                                            else "branch-top"
                                        ),
                                    )
                                if "*" in aa:
                                    nstyle["hz_line_color"] = "red"
            node.set_style(nstyle)

        ts = ete3.TreeStyle()
        ts.scale = scale
        ts.branch_vertical_margin = branch_margin
        ts.show_leaf_name = False
        ts.rotation = 90
        ts.draw_aligned_faces_as_table = False
        ts.allow_face_overlap = True
        ts.layout_fn = my_layout
        ts.show_scale = True
        ts.show_branch_support = show_support
        # if we labelled seqs, let's also write the alignment out so we have
        # the sequences (including of internal nodes)
        if idlabel:
            aln = MultipleSeqAlignment([])
            for node in tree_copy.traverse():
                aln.append(
                    SeqRecord(
                        Seq(str(node.sequence)),
                        id=str(node.name),
                        description=f"abundance={node.abundance}",
                    )
                )
            AlignIO.write(
                aln, open(os.path.splitext(outfile)[0] + ".fasta", "w"), "fasta"
            )
        return tree_copy.render(outfile, tree_style=ts)

    def feature_colormap(
        self,
        feature: str,
        cmap: str = "viridis",
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        scale: str = "linear",
        **kwargs,
    ) -> Dict[str, str]:
        r"""Generate a colormap based on a continuous tree feature.

        Args:
            feature: feature name (all nodes in tree attribute must have this feature)
            cmap: any `matplotlib color palette name <https://matplotlib.org/stable/gallery/color/colormap_reference.html>`_
            vmin: minimum value for colormap (default to minimum of the feature over the tree)
            vmax: maximum value for colormap (default to maximum of the feature over the tree)
            scale: ``linear`` (default), ``log``, or ``symlog`` (must also provide ``linthresh`` kwarg)
            kwargs: additional keyword arguments for scale transformation

        Returns:
            Dictionary of node names to hex color strings, which may be used as the colormap in :meth:`gctree.CollapsedTree.render`
        """
        cmap = mp.cm.get_cmap(cmap)

        if vmin is None:
            vmin = np.nanmin([getattr(node, feature) for node in self.tree.traverse()])
        if vmax is None:
            vmax = np.nanmax([getattr(node, feature) for node in self.tree.traverse()])

        if scale == "linear":
            norm = mp.colors.Normalize(vmin=vmin, vmax=vmax)
        elif scale == "log":
            norm = mp.colors.LogNorm(vmin=vmin, vmax=vmax)
        elif scale == "symlog":
            norm = mp.colors.SymLogNorm(vmin=vmin, vmax=vmax, **kwargs)
        else:
            raise ValueError(f"unrecognize scale: {scale}")

        return {
            node.name: mp.colors.to_hex(cmap(norm(getattr(node, feature))))
            for node in self.tree.traverse()
        }

    def write(self, file_name: str):
        r"""Serialize to pickle file.

        Args:
            file_name: file name (.p suffix recommended)
        """
        with open(file_name, "wb") as f:
            pickle.dump(self, f)

    def newick(self, file_name: str):
        r"""Write to newick file.

        Args:
            file_name: file name (.nk suffix recommended)
        """
        self.tree.write(format=1, outfile=file_name, format_root_node=True)

    def compare(
        self, tree2: CollapsedTree, method: str = "identity"
    ) -> Union[bool, np.float64]:
        r"""Compare this tree to the other tree.

        Args:
            tree2: another object of this type
            method: comparison type (``identity``, ``MRCA``, or ``RF``)

        Returns:
            tree difference
        """
        if method == "identity":
            # we compare lists of seq, parent, abundance
            # return true if these lists are identical, else false
            list1 = sorted(
                (
                    node.sequence,
                    node.abundance,
                    node.up.sequence if node.up is not None else None,
                )
                for node in self.tree.traverse()
            )
            list2 = sorted(
                (
                    node.sequence,
                    node.abundance,
                    node.up.sequence if node.up is not None else None,
                )
                for node in tree2.tree.traverse()
            )
            return list1 == list2
        elif method == "MRCA":
            # matrix of hamming distance of common ancestors of taxa
            # takes a true and inferred tree as CollapsedTree objects
            taxa = [node.sequence for node in self.tree.traverse() if node.abundance]
            n_taxa = len(taxa)
            d = np.zeros(shape=(n_taxa, n_taxa))
            sum_sites = np.zeros(shape=(n_taxa, n_taxa))
            for i in range(n_taxa):
                nodei_true = self.tree.iter_search_nodes(sequence=taxa[i]).__next__()
                nodei = tree2.tree.iter_search_nodes(sequence=taxa[i]).__next__()
                for j in range(i + 1, n_taxa):
                    nodej_true = self.tree.iter_search_nodes(
                        sequence=taxa[j]
                    ).__next__()
                    nodej = tree2.tree.iter_search_nodes(sequence=taxa[j]).__next__()
                    MRCA_true = self.tree.get_common_ancestor(
                        (nodei_true, nodej_true)
                    ).sequence
                    MRCA = tree2.tree.get_common_ancestor((nodei, nodej)).sequence
                    d[i, j] = gctree.utils.hamming_distance(MRCA_true, MRCA)
                    sum_sites[i, j] = len(MRCA_true)
            return d.sum() / sum_sites.sum()
        elif method == "RF":
            tree1_copy = self.tree.copy(method="deepcopy")
            tree2_copy = tree2.tree.copy(method="deepcopy")
            for treex in (tree1_copy, tree2_copy):
                for node in list(treex.traverse()):
                    if node.abundance > 0:
                        child = ete3.TreeNode()
                        child.add_feature("sequence", node.sequence)
                        node.add_child(child)
            return tree1_copy.robinson_foulds(
                tree2_copy,
                attr_t1="sequence",
                attr_t2="sequence",
                unrooted_trees=True,
            )[0]
        else:
            raise ValueError("invalid distance method: " + method)

    def _get_split(self, node: ete3.TreeNode) -> Tuple[Set, Set]:
        r"""Return the bipartition resulting from clipping this node's edge
        above.

        Args:
            node: tree node

        Returns:
            A tuple of two sets
        """
        if node.get_tree_root() != self.tree:
            raise ValueError("node not found")
        if node == self.tree:
            raise ValueError("this node is the root (no split above)")
        parent = node.up
        taxa1 = []
        for node2 in node.traverse():
            if node2.abundance > 0 or node2 == self.tree:
                if isinstance(node2.name, str):
                    taxa1.append(node2.name)
                else:
                    taxa1.extend(node2.name)
        taxa1 = set(taxa1)
        node.detach()
        taxa2 = []
        for node2 in self.tree.traverse():
            if node2.abundance > 0 or node2 == self.tree:
                if isinstance(node2.name, str):
                    taxa2.append(node2.name)
                else:
                    taxa2.extend(node2.name)
        taxa2 = set(taxa2)
        parent.add_child(node)
        assert taxa1.isdisjoint(taxa2)
        assert taxa1.union(taxa2) == set(
            (
                name
                for node in self.tree.traverse()
                if node.abundance > 0 or node == self.tree
                for name in ((node.name,) if isinstance(node.name, str) else node.name)
            )
        )
        return tuple(sorted([taxa1, taxa2]))

    @staticmethod
    def _split_compatibility(split1, split2):
        diff = split1[0].union(split1[1]) ^ split2[0].union(split2[1])
        if diff:
            raise ValueError(
                "splits do not cover the same taxa\n" f"\ttaxa not in both: {diff}"
            )
        for partition1 in split1:
            for partition2 in split2:
                if partition1.isdisjoint(partition2):
                    return True
        return False

    def support(
        self,
        bootstrap_trees_list: List[CollapsedTree],
        weights: Optional[List[np.float64]] = None,
        compatibility: bool = False,
    ):
        r"""Compute support from a list of bootstrap :class:`CollapsedTree`
        objects, and add to tree attibute.

        Args:
            bootstrap_trees_list: List of trees
            weights: weights for each tree, perhaps for weighting parsimony degenerate trees
            compatibility: counts trees that don't disconfirm the split.
        """
        for node in self.tree.get_descendants():
            split = self._get_split(node)
            support = 0
            compatibility_ = 0
            for i, tree in enumerate(bootstrap_trees_list):
                compatible = True
                supported = False
                for boot_node in tree.tree.get_descendants():
                    boot_split = tree._get_split(boot_node)
                    if (
                        compatibility
                        and compatible
                        and not self._split_compatibility(split, boot_split)
                    ):
                        compatible = False
                    if not compatibility and not supported and boot_split == split:
                        supported = True
                if supported:
                    support += weights[i] if weights is not None else 1
                if compatible:
                    compatibility_ += weights[i] if weights is not None else 1
            node.support = compatibility_ if compatibility else support

    def local_branching(
        self, tau=1, tau0=1, infinite_root_branch=True, nan_root_lbr=False
    ):
        r"""Add local branching statistics (Neher et al. 2014) as tree node
        features to the ETE tree attribute. After execution, all nodes will
        have new features ``LBI`` (local branching index) and ``LBR`` (local
        branching ratio, below Vs above the node)

        Args:
            tau: decay timescale for exponential filter
            tau0: effective branch length for branches with zero mutations
            infinite_root_branch: calculate assuming the root node has an infinite branch
            nan_root_lbr: replace the root LBR value with ``np.nan``
        """
        # the fixed integral contribution for clonal cells indicated by abundance annotations
        clone_contribution = tau * (1 - np.exp(-tau0 / tau))

        # post-order traversal to populate downward integrals for each node
        for node in self.tree.traverse(strategy="postorder"):
            if node.is_leaf():
                node.LB_down = {
                    node: (
                        node.abundance * clone_contribution if node.abundance > 1 else 0
                    )
                }
            else:
                node.LB_down = {node: node.abundance * clone_contribution}
                for child in node.children:
                    node.LB_down[child] = tau * (
                        1 - np.exp(-child.dist / tau)
                    ) + np.exp(-child.dist / tau) * sum(child.LB_down.values())

        # pre-order traversal to populate upward integral for each node
        for node in self.tree.traverse(strategy="preorder"):
            if node.is_root():
                # integral corresponding to infinite branch above root node
                node.LB_up = tau if infinite_root_branch else 0
            else:
                node.LB_up = tau * (1 - np.exp(-node.dist / tau)) + np.exp(
                    -node.dist / tau
                ) * (
                    node.up.LB_up
                    + sum(
                        node.up.LB_down[message]
                        for message in node.up.LB_down
                        if message != node
                    )
                )

        # finally, compute LBI (LBR) as the sum (ratio) of downward and upward integrals at each node
        for node in self.tree.traverse():
            node_LB_down_total = sum(node.LB_down.values())
            node.LBI = node_LB_down_total + node.LB_up
            node.LBR = node_LB_down_total / node.LB_up

        if nan_root_lbr:
            self.tree.LBR = np.nan


def _requires_dag(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._forest is None:
            raise NotImplementedError(
                f"CollapsedForest class method {func.__name__} may only be called on instances "
                "created from lists of ete3 trees. Instances created by simulation are not supported."
            )
        return func(self, *args, **kwargs)

    return wrapper


class CollapsedForest:
    r"""A collection of trees.

    We can intialize with a list of trees, each an instance of :class:`ete3.Tree` or :class:`CollapsedTree`, or we can simulate the forest later.

    Attributes:
        n_trees: number of trees in forest
        parameters: fit branching process parameters, if mle has been run, otherwise None

    Args:
        forest: list of :class:`ete3.Tree`
    """

    def __init__(
        self,
        forest: Optional[List[Union[CollapsedTree, ete3.Tree]]] = None,
    ):
        if forest is not None:
            if len(forest) == 0:
                raise ValueError("passed empty tree list")
            if isinstance(forest[0], CollapsedTree):
                forest = [ctree.tree for ctree in forest]
            elif not isinstance(forest[0], ete3.Tree):
                raise ValueError(
                    "If provided, `forest` argument should contain a list of ete3 trees or CollapsedTrees."
                )
            # Collect stats for validation
            model_tree = forest[0].copy()
            leaf_seqs = {node.sequence: node.name for node in model_tree.get_leaves()}
            root_seq = model_tree.sequence
            model_tree = disambiguate(model_tree)
            # include root in counts, since deduplicate will never let observed
            # root be a leaf.
            counts = {node.name: node.abundance for node in model_tree.iter_leaves()}
            counts[model_tree.name] = model_tree.abundance
            self._validation_stats = {
                "counts": counts,
                "root": model_tree.name,
                "parsimony_score": sum([node.dist for node in model_tree.traverse()]),
                "root_seq": root_seq,
            }
            if not any(_is_ambiguous(key) for key in leaf_seqs):
                self._validation_stats["leaf_seqs"] = leaf_seqs
            # Making this a private variable so that trying to access forest
            # attribute as before won't just give a confusing type or
            # attribute error.
            self._forest = _make_dag(forest)
            self.n_trees = self._forest.count_trees()
        else:
            self._forest = None
            self.n_trees = 0
            self._validation_stats = None
        self._cm_countlist = None
        self._ctrees = None
        self.parameters = None
        self.is_isotyped = False

    def simulate(self, p: np.float64, q: np.float64, n_trees: int):
        r"""Simulate a forest of collapsed trees. Overwrites existing forest
        attribute.

        Args:
            p: branching probability
            q: mutation probability
            n_trees: number of trees
        """
        self._forest = None
        self._ctrees = [CollapsedTree() for _ in range(n_trees)]
        self.n_trees = n_trees
        for tree in self._ctrees:
            tree.simulate(p, q)
        self._cm_countlist = tuple(
            coll.Counter([tree._cm_counts for tree in self._ctrees]).items()
        )

    @np.errstate(all="raise")
    def ll(
        self,
        p: np.float64,
        q: np.float64,
        marginal: bool = False,
    ) -> Tuple[np.float64, np.ndarray]:
        r"""Log likelihood of branching process parameters :math:`(p, q)` given
        tree topologies :math:`T_1, \dots, T_n` and corresponding genotype
        abundances vectors :math:`A_1, \dots, A_n` for each of :math:`n` trees
        in the forest.

        If ``marginal=False`` (the default), compute the joint log likelihood

        .. math::
            \ell(p, q; T, A) = \sum_{i=1}^n\log\mathbb{P}(T_i, A_i \mid p, q),

        otherwise compute the marginal log likelihood

        .. math::
            \ell(p, q; T, A) = \log\left(\sum_{i=1}^n\mathbb{P}(T_i, A_i \mid p, q)\right).

        Args:
            p: branching probability
            q: mutation probability
            marginal: compute the marginal likelihood over trees, otherwise compute the joint likelihood of trees

        Returns:
            Log branching process likelihood :math:`\ell(p, q; T, A)` and its gradient :math:`\nabla\ell(p, q; T, A)`
        """
        if self._cm_countlist is None:
            if self._forest is not None:
                cmcount_dagfuncs = _cmcounter_dagfuncs()

                def to_tuple(mset):
                    # When there's unobserved root unifurcation, augment with
                    # pseudocount.
                    if (0, 1) in mset:
                        assert mset[(0, 1)] == 1
                        mset = mset - {(0, 1)} + {(1, 1)}
                    return tuple(mset.items())

                cmcounters = self._forest.weight_count(**cmcount_dagfuncs)
                self._cm_countlist = [
                    (to_tuple(mset), mult) for mset, mult in cmcounters.items()
                ]
                # an iterable containing tuples `(cm_counts, mult)`, where `cm_counts`
                # is an iterable describing a tree, and `mult` is the number of trees in the forest
                # which can be described with the iterable `cm_counts`.
            elif self._ctrees is not None:
                self._cm_countlist = tuple(
                    coll.Counter([tree._cm_counts for tree in self._ctrees]).items()
                )
            else:
                raise ValueError("forest data must be defined to compute likelihood")

        terms = [
            [_lltree(cmcounts, p, q), count] for cmcounts, count in self._cm_countlist
        ]
        ls = np.array([term[0][0] for term in terms])
        grad_ls = np.array([term[0][1] for term in terms])
        # This can be done ahead of time
        count_ls = np.array([term[1] for term in terms])
        if marginal:
            # we need to find the smallest derivative component for each
            # coordinate, then subtract off to get positive things to logsumexp
            grad_l = []
            for j in range(len((p, q))):
                i_prime = grad_ls[:, j].argmin()
                b = (grad_ls[:, j] - grad_ls[i_prime, j]) * count_ls
                # believe it or not, logsumexp can't handle 0 in b
                # when np.seterr(underflow='raise') on newer processors:
                if all(b == 0):
                    # logsumexp can't take empty list, but if b is 0, then
                    # logsumexp should be -inf, whose exponent is 0:
                    res = 0
                else:
                    res = np.exp(
                        scs.logsumexp((ls - ls[i_prime])[b != 0], b=b[b != 0])
                        - scs.logsumexp(ls - ls[i_prime], b=count_ls)
                    )
                grad_l.append(grad_ls[i_prime, j] + res)
            # count_ls shouldn't have any zeros in it...
            # using math.log instead of np.log is essential because np.log
            # doesn't work on large integers > 2**64 :eyeroll:
            return (
                -math.log(count_ls.sum()) + scs.logsumexp(ls, b=count_ls)
            ), np.array(grad_l)
        else:
            return (ls * count_ls).sum(), np.array(
                [(grad_ls[:, 0] * count_ls).sum(), (grad_ls[:, 1] * count_ls).sum()]
            )

    def mle(self, **kwargs) -> Tuple[np.float64, np.float64]:
        r"""Maximum likelihood estimate of :math:`(p, q)`.

        .. math::
            (p, q) = \arg\max_{p,q\in [0,1]}\ell(p, q)

        Args:
            kwargs: keyword arguments passed along to the branching process likelihood :meth:`CollapsedForest.ll`

        Returns:
            Tuple :math:`(p, q)` with estimated branching probability and estimated mutation probability
        """
        self.parameters = _mle_helper(self.ll, **kwargs)
        return self.parameters

    @_requires_dag
    def filter_trees(  # noqa: C901
        self,
        ranking_coeffs: Optional[Sequence[float]] = None,
        mutability_file: Optional[str] = None,
        substitution_file: Optional[str] = None,
        ignore_isotype: bool = False,
        chain_split: Optional[int] = None,
        verbose: bool = False,
        outbase: str = "gctree.out",
        summarize_forest: bool = False,
        tree_stats: bool = False,
        branching_process_ranking_coeff: float = -1,
        use_old_mut_parsimony: bool = False,
    ) -> CollapsedForest:
        """Filter trees according to specified criteria.

        Trim the forest to minimize a linear
        combination of branching process likelihood, isotype parsimony score,
        context/mutability-based Poisson likelihood, and number of alleles, with coefficients
        provided in the argument ``ranking_coeffs`, in that order.

        Args:
            ranking_coeffs: A list or tuple of coefficients for prioritizing tree weights.
                The order of coefficients is: isotype parsimony score, context poisson likelihood,
                and number of alleles. A coefficient of ``-1`` will be applied to branching process
                likelihood by default, unless a different value is provided to the keyword argument
                `branching_process_ranking_coeff`. Trees are chosen to minimize this linear combination
                of tree weights, so weights for which larger values are more optimal (such as
                likelihoods) should have negative coefficients.
                If ranking_coeffs is not provided, trees will be ranked lexicographically
                by likelihood, then by other traits, in the same order.
            mutability_file: A mutability model
            substitution_file: A substitution model
            ignore_isotype: Ignore isotype parsimony when ranking. By default, isotype information added with
                :meth:``add_isotypes`` will be used to compute isotype parsimony, which is used in ranking.
            chain_split: The index at which non-adjacent sequences are concatenated, for calculating
                context-based Poisson likelihood.
            verbose: print information about trimming
            outbase: file name stem for a file with information for each tree in the DAG.
            summarize_forest: whether to write a summary of the forest to file `[outbase].forest_summary.log`
            tree_stats: whether to write stats for each tree in the forest to file `[outbase].tree_stats.log`
            branching_process_ranking_coeff: Ranking coefficient to use for branching process likelihood. Value
                is ignored unless `ranking_coeffs` argument is provided.
            use_old_mut_parsimony: Whether to use the deprecated 'mutability parsimony' instead of
                context-based poisson likelihood (only applicable if mutability and substitution files are
                provided.

        Returns:
            The trimmed forest, containing all optimal trees according to the specified criteria, and a tuple
            of data about the trees in that forest, with format (branching process likelihood, isotype parsimony,
            context-based Poisson likelihood, alleles).
        """
        dag = self._forest

        if ranking_coeffs:
            if len(ranking_coeffs) != 3:
                raise ValueError(
                    "If ranking_coeffs are provided to `filter_trees` method, a list of three values is expected."
                )
            coeffs = [branching_process_ranking_coeff] + list(ranking_coeffs)
            if sum(abs(c) for c in coeffs) == 0:
                raise ValueError(
                    "At least one value provided to ranking_coeffs or the value of branching_process_ranking_coeff must be nonzero."
                )
        else:
            coeffs = [1] * 4

        (
            nz_coeff_bplikelihood,
            nz_coeff_isotype_pars,
            nz_coeff_context,
            nz_coeff_alleles,
        ) = [val != 0 for val in coeffs]
        coeff_bplikelihood, coeff_isotype_pars, coeff_context, coeff_alleles = coeffs

        dag_filters = []
        if nz_coeff_bplikelihood:
            if self.parameters is None:
                self.mle(marginal=True)
            p, q = self.parameters
            ll_dagfuncs = _ll_genotype_dagfuncs(p, q)
            dag_filters.append((ll_dagfuncs, coeff_bplikelihood))
            if verbose:
                print(f"Branching process parameters to be used for ranking: {(p, q)}")
        if nz_coeff_isotype_pars and self.is_isotyped and (not ignore_isotype):
            # Check for missing isotype data in all but root node, and fake root-adjacent leaf node
            rootname = list(self._forest.dagroot.children())[0].attr["name"]
            if any(
                not node.attr["isotype"]
                for node in self._forest.preorder()
                if not node.is_root() and node.attr["name"] != rootname
            ):
                warnings.warn(
                    "Some isotype data seems to be missing. Isotype parsimony scores may be incorrect."
                )

            iso_funcs = _isotype_dagfuncs()
            dag_filters.append((iso_funcs, coeff_isotype_pars))
        if nz_coeff_context and mutability_file and substitution_file:
            if use_old_mut_parsimony:
                mut_funcs = _mutability_dagfuncs(
                    mutability_file=mutability_file,
                    substitution_file=substitution_file,
                    splits=[] if chain_split is None else [chain_split],
                )
            else:
                mut_funcs = _context_poisson_likelihood_dagfuncs(
                    mutability_file=mutability_file,
                    substitution_file=substitution_file,
                    splits=[] if chain_split is None else [chain_split],
                )
            dag_filters.append((mut_funcs, coeff_context))
        if nz_coeff_alleles:
            allele_funcs = _allele_dagfuncs()
            dag_filters.append((allele_funcs, coeff_alleles))

        combined_dag_filter = functools.reduce(
            lambda x, y: x + y, (dag_filter for dag_filter, _ in dag_filters)
        )
        if ranking_coeffs:
            if len(ranking_coeffs) != 3:
                raise ValueError(
                    "If ranking_coeffs are provided to `filter_trees` method, a list of three values is expected."
                )
            for dag_filter, coeff in dag_filters:
                if dag_filter.optimal_func == max and coeff > 0:
                    warnings.warn(
                        f"Higher values for {dag_filter.weight_funcs.name} are generally better, but the "
                        "provided ranking coefficient is positive, so trees with lower values will be preferred."
                    )
                if dag_filter.optimal_func == min and coeff < 0:
                    warnings.warn(
                        f"Lower values for {dag_filter.weight_funcs.name} are generally better, but the "
                        "provided ranking coefficient is negative, so trees with higher values will be preferred."
                    )

            filtered_coefficients = [coeff for _, coeff in dag_filters]

            def linear_combinator(weighttuple):
                return sum(
                    [
                        priority * float(weight)
                        for priority, weight in zip(filtered_coefficients, weighttuple)
                    ]
                )

            old_edge_func = combined_dag_filter["edge_weight_func"]
            ranking_dag_filter = hdag.utils.HistoryDagFilter(
                hdag.utils.AddFuncDict(
                    {
                        "start_func": lambda n: 0,
                        "edge_weight_func": lambda n1, n2: linear_combinator(
                            old_edge_func(n1, n2)
                        ),
                        "accum_func": sum,
                    }
                ),
                min,
                ordering_name="LinearCombination",
            )
            ranking_description = (
                "Ranking trees to minimize a linear combination of "
                + " + ".join(
                    str(coeff) + "(" + fl.weight_funcs.name + ")"
                    for fl, coeff in dag_filters
                )
            )
        else:
            ranking_dag_filter = combined_dag_filter
            ranking_description = "Ranking trees to " + " then ".join(
                opt_name[:3] + "imize " + ord_name
                for (opt_name, _), ord_name in zip(
                    ranking_dag_filter.ordering_names,
                    ranking_dag_filter.weight_funcs.names,
                )
            )
        if verbose:
            print(ranking_description)

        def print_stats(statlist, title, file=None, suppress_score=False):
            show_score = ranking_coeffs and not suppress_score

            def reformat(field, n=10):
                if isinstance(field, int):
                    return format(field, "<" + str(n))
                else:
                    return f"{field:{n}.{n}}"

            print("\n" + title + ":", file=file)
            statstring = "\t".join(
                tuple(
                    reformat(dfilter.weight_funcs.name, n=14)
                    for dfilter, _ in dag_filters
                )
            )
            print(
                f"tree     \t{statstring}" + ("\ttreescore" if show_score else ""),
                file=file,
            )
            for j, best_weighttuple in enumerate(statlist, 1):
                statstring = "\t".join(reformat(it) for it in best_weighttuple)
                print(
                    f"{j:<10}\t{statstring}"
                    + (
                        f"\t{reformat(linear_combinator(best_weighttuple))}"
                        if show_score
                        else ""
                    ),
                    file=file,
                )

        trimdag = dag[ranking_dag_filter]

        best_weighttuple = trimdag.optimal_weight_annotate(
            **combined_dag_filter,
        )

        if verbose:
            print_stats([best_weighttuple], "Stats for optimal trees")

        if summarize_forest:
            with open(outbase + ".forest_summary.log", "w") as fh:
                independent_best = []
                for dfilter, _ in dag_filters:
                    tempdag = dag.copy()
                    min_val, max_val = tempdag.weight_range_annotate(**dfilter)
                    opt_weight = tempdag.trim_optimal_weight(**dfilter)
                    independent_best.append(opt_weight)
                    fh.write(
                        f"\nOverall {dfilter.weight_funcs.name} range {min_val} to {max_val}."
                        f"\nAmong trees with {dfilter.optimal_func.__name__} {dfilter.weight_funcs.name} of: {opt_weight}\n"
                    )
                    for indfilter, _ in dag_filters:
                        if indfilter.weight_funcs.name != dfilter.weight_funcs.name:
                            minval, maxval = tempdag.weight_range_annotate(
                                **indfilter.weight_funcs
                            )
                            fh.write(
                                f"\t{indfilter.weight_funcs.name} range: {minval} to {maxval}\n"
                            )
                print("\n", file=fh)
                print_stats(
                    [
                        [
                            stat - best
                            for stat, best in zip(best_weighttuple, independent_best)
                        ]
                    ],
                    "Highest ranked tree: loss from best value",
                    file=fh,
                    suppress_score=True,
                )

        if tree_stats:
            dag_ls = list(dag.weight_count(**combined_dag_filter).elements())
            # To clear _dp_data fields of their large cargo
            dag.optimal_weight_annotate(edge_weight_func=lambda n1, n2: 0)
            if ranking_coeffs:
                minfunckey = linear_combinator
            else:
                minfunckey = ranking_dag_filter.optimal_func
            dag_ls.sort(key=minfunckey)

            df = pd.DataFrame(dag_ls, columns=combined_dag_filter.weight_funcs.names)
            df.to_csv(outbase + ".tree_stats.csv")
            df["set"] = ["all_trees"] * len(df)
            bestdf = pd.DataFrame(
                [best_weighttuple], columns=combined_dag_filter.weight_funcs.names
            )
            bestdf["set"] = ["best_tree"]
            toplot_df = pd.concat([df, bestdf], ignore_index=True)
            pplot = sns.pairplot(
                toplot_df.drop(["Alleles"], errors="ignore"),
                hue="set",
                diag_kind="hist",
            )
            pplot.savefig(outbase + ".tree_stats.pairplot.pdf")

        return (self._trimmed_self(trimdag), best_weighttuple)

    def likelihood_rankplot(self, outbase, p, q, img_type="svg"):
        """Save a rank plot of likelihoods to the file
        `[outbase].inference.likelihood_rank.[img_type]`."""
        ll_dagfuncs = _ll_genotype_dagfuncs(p, q)
        if self._forest is not None:
            dag_l = list(
                float(ll) for ll in self._forest.weight_count(**ll_dagfuncs).elements()
            )
        else:
            dag_l = [ctree.ll(p, q)[0] for ctree in self]
        dag_l.sort(key=lambda n: -n)
        plt.figure(figsize=(6.5, 2))
        try:
            plt.plot(np.exp(dag_l), "ko", clip_on=False, markersize=4)
            plt.ylabel("gctree likelihood")
            plt.yscale("log")
            plt.ylim([None, 1.1 * max(np.exp(dag_l))])
        except FloatingPointError:
            plt.plot(dag_l, "ko", clip_on=False, markersize=4)
            plt.ylabel("gctree log-likelihood")
            plt.ylim([None, 1.1 * max(dag_l)])
        plt.xlabel("parsimony tree")
        plt.xlim([-1, len(dag_l)])
        plt.tick_params(axis="y", direction="out", which="both")
        plt.tick_params(
            axis="x", which="both", bottom="off", top="off", labelbottom="off"
        )
        plt.savefig(outbase + ".inference.likelihood_rank." + img_type)

    @_requires_dag
    def n_topologies(self) -> int:
        """Count the number of topology classes, ignoring internal node
        sequences."""
        return self._forest.count_topologies(collapse_leaves=True)

    @_requires_dag
    def iter_topology_classes(self):
        """Sort trees by topology class.

        Returns:
            A generator of CollapsedForest objects, each containing trees with the same topology,
                ignoring internal node labels. CollapsedForests will be yielded in reverse-order
                of the number of trees in each topology class, so that each CollapsedForest will
                contain at least as many trees as the one that follows.
        """
        topology_dagfuncs = hdag.utils.make_newickcountfuncs(
            internal_labels=False, collapse_leaves=True
        )
        ctopologies = list(self._forest.weight_count(**topology_dagfuncs).items())
        # yield classes with most members first
        ctopologies.sort(key=lambda t: -t[1])
        for ctopology, count in ctopologies:
            tdag = self._forest.copy()
            tdag.trim_topology(ctopology, collapse_leaves=True)
            yield self._trimmed_self(tdag)

    @_requires_dag
    def add_isotypes(
        self,
        isotypemap: Optional[Mapping[str, str]] = None,
        isotypemap_file: Optional[str] = None,
        idmap: Optional[Mapping[str, Set[str]]] = None,
        idmap_file: Optional[str] = None,
        isotype_names: Optional[Sequence[str]] = None,
    ):
        """Adds isotype annotations, including inferred ancestral isotypes, to
        all nodes in stored trees."""
        self.is_isotyped = True

        iso_funcs = _isotype_annotation_dagfuncs(
            isotypemap_file=isotypemap_file,
            isotypemap=isotypemap,
            idmap_file=idmap_file,
            idmap=idmap,
            isotype_names=isotype_names,
        )

        def optimal_func(weightlist):
            # all weights (frozendicts containing isotypes) are equal
            return weightlist[0]

        self._forest.optimal_weight_annotate(**iso_funcs, optimal_func=optimal_func)
        for node in self._forest.preorder(skip_root=True):
            node.attr["isotype"] = node._dp_data

    def sample_tree(self) -> CollapsedTree:
        """Sample a random CollapsedTree from the forest."""
        if self._ctrees is not None:
            return random.choice(self._ctrees)
        elif self._forest is not None:
            return self._clade_tree_to_ctree(self._forest.sample())
        else:
            raise ValueError("Cannot sample trees from an empty forest")

    def _trimmed_self(self, dag):
        """Create a new CollapsedForest object from a subset of the current
        one's history DAG, properly initializing validation_stats and
        n_trees."""
        newforest = CollapsedForest()
        newforest._validation_stats = self._validation_stats.copy()
        newforest.n_trees = dag.count_trees()
        newforest._forest = dag
        newforest.parameters = self.parameters
        return newforest

    def _clade_tree_to_ctree(
        self,
        clade_tree: hdag.HistoryDag,
    ) -> CollapsedTree:
        """Create and validate :meth:`CollapsedTree` object from tree-shaped
        history DAG.

        Uses self._validation_stats dictionary, if available, for validation. Dictionary keys are:
            root: The expected root name for the resulting tree (for validation)
            counts: Expected node abundances for each node name (for validation)
            parsimony_score: Expected Hamming parsimony score (for validation)
            root_seq: Expected root sequence (for validation)
            leaf_seqs: Expected leaf names, keyed by leaf sequences (for validation)

        Args:
            clade_tree: A tree-shaped history DAG, like that returned by :meth:`historydag.HistoryDag.sample`

        Returns:
            :meth:`CollapsedTree` object matching the topology of ``clade_tree``, but fully collapsed.
        """
        # Allow only leaf nodes to have nonzero abundance, so that ctree
        # collapsing takes care of all uncollapsed leaves. This works for
        # observed root because all trees have added pseudo leaf with root name
        # and abundance.
        # original_ids is a set of original leaf node ids, or empty if node is
        # not a leaf.
        etetree = clade_tree.to_ete(
            name_func=lambda n: n.attr["name"],
            features=["sequence"],
            feature_funcs={
                "abundance": lambda n: n.label.abundance if n.is_leaf() else 0,
                "isotype": lambda n: n.attr["isotype"],
                "original_ids": lambda n: n.attr["original_ids"],
            },
        )

        ctree = CollapsedTree(etetree)

        # Fix internal node names to be unique, and verify
        # The maps from nodes to names and nodes to sequences are bijections
        n_nodes = 0
        names = set()
        seqs = set()
        for node in ctree.tree.traverse():
            n_nodes += 1
            names.add(node.name)
            seqs.add(node.sequence)
        n_names, n_seqs = len(names), len(seqs)
        if not (n_nodes == n_names and n_names == n_seqs):
            raise RuntimeError(
                "Multiple sequences with the same name, or multiple"
                "names for the same sequence, observed in collapsed tree."
            )

        # Here can do some validation on the tree:
        if self._validation_stats is not None:
            # root name:
            if self._validation_stats["root"] != ctree.tree.name:
                raise RuntimeError(
                    f"collapsed tree should have root name '{self._validation_stats['root']}' but has instead {ctree.tree.name}"
                )
            # counts:
            counts = self._validation_stats["counts"]
            for node in etetree.iter_leaves():
                assert (
                    sum(counts[og_id] for og_id in node.original_ids) == node.abundance
                )
                assert node.name in node.original_ids
            tree_abundance = 0
            for node in ctree.tree.traverse():
                tree_abundance += node.abundance
                if node.name in counts:
                    assert (
                        sum(counts[og_id] for og_id in node.original_ids)
                        == node.abundance
                    )
                    assert node.name in node.original_ids
                else:
                    assert node.abundance == 0
            assert tree_abundance == sum(counts.values())

            # unnamed_seq issue:
            for node in ctree.tree.traverse():
                if node.name == "unnamed_seq":
                    raise RuntimeError("Some node names are missing")

            # Parsimony:
            if self._validation_stats["parsimony_score"] != sum(
                [
                    gctree.utils.hamming_distance(node.up.sequence, node.sequence)
                    for node in ctree.tree.iter_descendants()
                ]
            ):
                raise RuntimeError(
                    "History DAG tree parsimony score does not match parsimony score provided"
                )
            # Root sequence is a possible disambiguation of root:
            if any(
                base not in gctree.utils.ambiguous_dna_values[ambig_base]
                for base, ambig_base in zip(
                    ctree.tree.sequence, self._validation_stats["root_seq"]
                )
            ):
                raise RuntimeError(
                    "History DAG root node sequence does not match root sequence provided\n"
                    "found: " + ctree.tree.sequence + "\n"
                    "expected: " + self._validation_stats["root_seq"]
                )
            # Leaf names:
            if "leaf_seqs" in self._validation_stats:
                # Will be intentionally missing if observed sequences had
                # ambiguities.
                leaf_seqs = self._validation_stats["leaf_seqs"]
                # A dictionary of leaf sequences to leaf names
                observed_set = {
                    node for node in ctree.tree.iter_descendants() if node.abundance > 0
                }
                for node in observed_set:
                    if not node.is_root() and leaf_seqs[node.sequence] != node.name:
                        raise RuntimeError(
                            "History DAG tree leaf names don't match sequences"
                        )
                observed_seqs = {node.sequence for node in observed_set}
                nonroot_observed_seqs = observed_seqs - {ctree.tree.sequence}
                nonroot_leaf_seqs = set(leaf_seqs.keys()) - {ctree.tree.sequence}
                if nonroot_leaf_seqs != nonroot_observed_seqs:
                    raise RuntimeError(
                        "Observed nonroot sequences in history DAG tree don't match "
                        "observed nonroot sequences passed in leaf_seqs."
                    )
        else:
            warnings.warn("No validation was performed on tree")

        return ctree

    def __repr__(self):
        r"""Return a string representation for printing."""
        return f"n_trees = {self.n_trees}\n" "\n".join([str(tree) for tree in self])

    def __iter__(self):
        if self._ctrees is not None:
            yield from self._ctrees
        elif self._forest is not None:
            for cladetree in self._forest.get_trees():
                yield self._clade_tree_to_ctree(cladetree)
        else:
            yield from ()

    def __getstate__(self):
        # Avoid pickling large cached abundance data.
        # hDAG also defines its own getstate.
        d = self.__dict__.copy()
        d["_cm_countlist"] = None
        return d


def _mle_helper(
    ll: Callable[[np.float64, np.float64], Tuple[np.float64, np.ndarray]], **kwargs
) -> Tuple[np.float64, np.float64]:
    # initialization
    x_0 = (0.5, 0.5)
    bounds = ((1e-6, 1 - 1e-6), (1e-6, 1 - 1e-6))

    def f(x):
        """Negative log branching process likelihood."""
        return tuple(-y for y in ll(*x, **kwargs))

    grad_check = sco.check_grad(lambda x: f(x)[0], lambda x: f(x)[1], x_0)
    if grad_check > 1e-3:
        warnings.warn(
            f"gradient mismatches finite difference approximation by {grad_check}",
            RuntimeWarning,
        )
    result = sco.minimize(
        f,
        jac=True,
        x0=x_0,
        method="L-BFGS-B",
        options={"ftol": 1e-10},
        bounds=bounds,
    )
    # update params if None and optimization successful
    if not result.success:
        warnings.warn("optimization not sucessful, " + result.message, RuntimeWarning)
    return result.x[0], result.x[1]


def _lltree(cm_counts, p: np.float64, q: np.float64) -> Tuple[np.float64, np.ndarray]:
    r"""Log likelihood of branching process parameters :math:`(p, q)`

    .. math::
        \ell(p, q; T, A) = \log\mathbb{P}(T, A \mid p, q)

    Args:
        cm_counts: an iterable containing tuples `((c, m), n)` where `n` is the number of nodes
            in the tree with abundance `c` and `m` mutant clades
        p: branching probability
        q: mutation probability
    Returns:
        Log likelihood :math:`\ell(p, q; T, A)` and its gradient :math:`\nabla\ell(p, q; T, A)`
    """
    count_ls = [n for cm, n in cm_counts]
    cm_list = [cm for cm, n in cm_counts]
    # extract vector of function values and gradient components
    logf_data = [CollapsedTree._ll_genotype(c, m, p, q) for c, m in cm_list]
    logf = np.array([[x[0] * count_ls[i]] for i, x in enumerate(logf_data)]).sum()
    grad_ll_genotype = np.array(
        [x[1] * count_ls[i] for i, x in enumerate(logf_data)]
    ).sum(axis=0)
    return logf, grad_ll_genotype


def _is_ambiguous(sequence):
    return any(base not in gctree.utils.bases for base in sequence)


def _make_dag(trees, from_copy=True, quick_sankoff=False):
    """Build a history DAG from ambiguous or disambiguated trees, whose nodes
    have abundance, name, and sequence attributes."""
    # preprocess trees so they're acceptable inputs
    # Assume all trees have fixed root sequence and fixed leaf sequences

    if from_copy:
        trees = [tree.copy() for tree in trees]

    # disambiguate leaves: disambiguate each tree and transplant disambiguated
    # leaf sequences to tree with ambiguous internal sequences

    # add pseudo-leaf below root in all trees:
    # This is done first in case root is ambiguous, and gets merged with
    # another ambiguous leaf node after disambiguation.
    rootname = trees[0].name  # all root nodes must have the same name
    for tree in trees:
        newleaf = tree.add_child(name=tree.name, dist=0)
        newleaf.add_feature("sequence", tree.sequence)
        newleaf.add_feature("abundance", tree.abundance)

    if any(_is_ambiguous(leaf.sequence) for leaf in trees[0].iter_leaves()):
        warnings.warn(
            "Some observed sequences are ambiguous. A disambiguation consistent"
            " with each dnapars tree will be chosen arbitrarily. Many alternative"
            " disambiguated leaf sequences may be possible."
        )
        for tree in trees:
            for node in tree.iter_leaves():
                node.add_feature("original_ids", {node.name})
            disambig_tree = tree.copy()
            node_map = {
                d_node: o_node
                for d_node, o_node in zip(disambig_tree.traverse(), tree.traverse())
            }
            disambiguate(disambig_tree)

            # remove duplicate leaves, and adjust abundances
            leaf_seqs = {}
            to_delete = []
            for leaf in disambig_tree.iter_leaves():
                if leaf.sequence in leaf_seqs:
                    leaf_seqs[leaf.sequence].append(leaf)
                else:
                    leaf_seqs[leaf.sequence] = [leaf]
            for sequence, leaf_list in leaf_seqs.items():
                if len(leaf_list) > 1:
                    # Always choose root pseudo-leaf to represent nodes, if
                    # possible
                    ancestor = disambig_tree.get_common_ancestor(*leaf_list)
                    _leaf_list_names = {node.name: node for node in leaf_list}
                    if rootname in _leaf_list_names:
                        rep_node = _leaf_list_names.pop(rootname)
                    else:
                        rep_node = _leaf_list_names.pop(leaf_list[0].name)
                    to_delete = list(_leaf_list_names.values())
                    rep_node.abundance = sum(leaf.abundance for leaf in leaf_list)
                    rep_node.original_ids = {
                        seq_id for node in leaf_list for seq_id in node.original_ids
                    }
                    while to_delete:
                        for node in to_delete:
                            node_map[node].delete(prevent_nondicotomic=False)
                            node.delete(prevent_nondicotomic=False)
                        to_delete = [
                            leaf
                            for leaf in ancestor.iter_leaves()
                            if (leaf.sequence == sequence and leaf != rep_node)
                        ]
            # transplant leaf sequences and abundances:
            for node in disambig_tree.iter_leaves():
                node_map[node].abundance = node.abundance
                node_map[node].sequence = node.sequence
                node_map[node].original_ids = node.original_ids
            # remove some unifurcations
            to_delete = []
            for node in tree.iter_descendants():
                if len(node.children) == 1:
                    # this excludes leaves
                    to_delete.append(node)
            for node in to_delete:
                node.delete(prevent_nondicotomic=False)

    def trees_to_dag(trees):
        return hdag.history_dag_from_etes(
            trees,
            ["sequence"],
            label_functions={"abundance": lambda n: n.abundance if n.is_leaf() else 0},
            attr_func=lambda n: {
                "name": n.name,
                "original_ids": (
                    n.original_ids
                    if "original_ids" in n.features
                    else {n.name} if n.is_leaf() else set()
                ),
                "isotype": frozendict(),
            },
        )

    dag = trees_to_dag(trees)
    # If there are too many ambiguities at too many nodes, disambiguation will
    # hang. Need to have an alternative (disambiguate each tree before putting in dag):

    def test_explode_individually():
        try:
            if (
                dag.count_trees(expand_count_func=sequence_resolutions_count)
                / dag.count_trees()
                > 5000000
            ):
                return True
            else:
                return False
        except OverflowError:
            return True

    if test_explode_individually():
        warnings.warn(
            "Parsimony trees have too many ambiguities for disambiguation in all possible ways. "
            "Disambiguating trees individually. Gctree may find fewer parsimony trees."
        )
        distrees = [disambiguate(tree) for tree in trees]
        dag = trees_to_dag(distrees)
    dag.explode_nodes(expand_func=sequence_resolutions)
    # Look for (even) more trees:
    dag.add_all_allowed_edges(adjacent_labels=True)
    dag.trim_optimal_weight()
    for node in dag.preorder(skip_root=True):
        if node.label.abundance != 0 and not node.is_leaf():
            raise RuntimeError(
                f"An internal node {node.attr['name']} was found with nonzero abundance {node.label.abundance}"
            )
    dag.convert_to_collapsed()
    dag.recompute_parents()
    # Only leaf nodes have nonzero abundance now, so all internal edges are
    # collapsed by sequence. Now labels with correct abundances are placed on
    # leaf-adjacent nodes that will be collapsed by sequence.
    for node in dag.preorder(skip_root=True):
        if node.is_leaf():
            for parent in node.parents:
                if parent.label.sequence == node.label.sequence:
                    parent.label = node.label

    dag = hdag.sequence_dag.SequenceHistoryDag.from_history_dag(dag)
    pars_count = dag.hamming_parsimony_count()
    if len(pars_count) > 1:
        raise RuntimeError(
            f"History DAG parsimony search resulted in parsimony trees of unexpected weights:\n {pars_count}"
        )

    # names on internal nodes are all messed up from disambiguation step, we'll
    # fix them in CollapsedTree.__init__.
    return dag


def _cmcounter_dagfuncs():
    """Functions for accumulating frozen multisets of (c, m) pairs in trees in
    the DAG."""

    def edge_weight_func(n1, n2):
        if n2.is_leaf() and n1.label.sequence == n2.label.sequence:
            # Then this is a leaf-adjacent node with nonzero abundance
            return multiset.FrozenMultiset()
        else:
            m = len(n2.clades)
            if frozenset({n2.label}) in n2.clades:
                m -= 1
            return multiset.FrozenMultiset([(n2.label.abundance, m)])

    def accum_func(cmsetlist: List[multiset.FrozenMultiset]):
        st = multiset.FrozenMultiset()
        for cmset in cmsetlist:
            st += cmset
        return st

    return hdag.utils.AddFuncDict(
        {
            "start_func": lambda n: multiset.FrozenMultiset(),
            "edge_weight_func": edge_weight_func,
            "accum_func": accum_func,
        },
        name="cmcounters",
    )


def _ll_genotype_dagfuncs(p: np.float64, q: np.float64) -> hdag.utils.HistoryDagFilter:
    """Return functions for counting tree log branching process likelihood on
    the history DAG.

    For numerical consistency, we resort to the use of ``decimal.Decimal``.
    This is exactly for the purpose of solving the problem that float sum is
    sensitive to order of summation, while Decimal sum is not.

    Still, there seems to be some inconsistency (probably in the part of the computation done with floats)
    so we wrap the Decimal in a :class:`historydag.utils.FloatState` object, where the exposed
    float is a rounded version of the hidden Decimal state, which is used for actual computations.


    Args:
        p, q: branching process parameters

    Returns:
        A :meth:`historydag.utils.HistoryDagFilter` which may be passed as keyword arguments
        to :meth:`historydag.HistoryDag.weight_count`, :meth:`historydag.HistoryDag.trim_optimal_weight`,
        or :meth:`historydag.HistoryDag.optimal_weight_annotate`
        methods to trim or annotate a :meth:`historydag.HistoryDag` according to branching process likelihood.
        Weight format is ``decimal.Decimal``.
    """

    def edge_weight_ll_genotype(n1: hdag.HistoryDagNode, n2: hdag.HistoryDagNode):
        """The _ll_genotype weight of the target node, unless it should be
        collapsed, then 0.

        Expects DAG to have abundances added so that each node has
        abundance feature on label.
        """
        if n2.is_leaf() and n2.label.sequence == n1.label.sequence:
            return hdag.utils.FloatState(0.0, state=Decimal(0.0))
        else:
            m = len(n2.clades)
            # Check if this edge should be collapsed, and reduce mutant descendants
            if frozenset({n2.label}) in n2.clades:
                m -= 1
            c = n2.label.abundance
            if n1.is_root() and c == 0 and m == 1:
                # Add pseudocount for unobserved root unifurcation
                c = 1
            res = Decimal(CollapsedTree._ll_genotype(c, m, p, q)[0])
            return hdag.utils.FloatState(float(round(res, 8)), state=res)

    def accum_func(weightlist):
        res = sum(weight.state for weight in weightlist)
        return hdag.utils.FloatState(float(round(res, 8)), state=res)

    return hdag.utils.HistoryDagFilter(
        hdag.utils.AddFuncDict(
            {
                "start_func": lambda n: hdag.utils.FloatState(0.0, state=Decimal(0)),
                "edge_weight_func": edge_weight_ll_genotype,
                "accum_func": accum_func,
            },
            name="LogBPLikelihood",
        ),
        max,
    )


def _allele_dagfuncs() -> hdag.utils.HistoryDagFilter:
    """Return functions for filtering trees in a history DAG by allele count.

    The number of alleles in a tree is the number of unique sequences observed on nodes of that tree.

    Returns:
        A :meth:`historydag.utils.AddFuncDict` which may be passed as keyword arguments
        to :meth:`historydag.HistoryDag.weight_count`, :meth:`historydag.HistoryDag.trim_optimal_weight`,
        or :meth:`historydag.HistoryDag.optimal_weight_annotate`
        methods to trim or annotate a :meth:`historydag.HistoryDag` according to allele count.
        Weight format is ``int``.
    """
    return hdag.utils.HistoryDagFilter(
        hdag.utils.AddFuncDict(
            {
                "start_func": lambda n: 0,
                "edge_weight_func": lambda n1, n2: n1.label != n2.label,
                "accum_func": sum,
            },
            name="Alleles",
        ),
        min,
    )



r"""Utility functions."""

from functools import wraps, reduce
import Bio.Data.IUPACData
import operator
from typing import Sequence, Any

Multiplicable = Any

bases = "AGCT-"
ambiguous_dna_values = Bio.Data.IUPACData.ambiguous_dna_values.copy()
ambiguous_dna_values.update({"?": "GATC-", "-": "-"})
ambiguous_dna_keys = {frozenset(val): key for key, val in ambiguous_dna_values.items()}


def _check_distance_arguments(distance):
    @wraps(distance)
    def new_distance(seq1: str, seq2: str, *args, **kwargs):
        if len(seq1) != len(seq2):
            raise ValueError(
                f"sequences must have equal length, got {len(seq1)} and {len(seq2)}"
            )
        return distance(seq1, seq2, *args, **kwargs)

    return new_distance


@_check_distance_arguments
def hamming_distance(seq1: str, seq2: str) -> int:
    r"""Hamming distance between two sequences of equal length.

    Args:
        seq1: sequence 1
        seq2: sequence 2
    """
    return sum(x != y for x, y in zip(seq1, seq2))


def product(
    factors: Sequence[Multiplicable], f=operator.mul, identity=1
) -> Multiplicable:
    return reduce(f, factors, identity)



import gctree.branching_processes as bp
import gctree.phylip_parse as pp
import gctree.utils as utils
import gctree.mutation_model as mm
from math import log

import numpy as np
from multiset import FrozenMultiset
from oldcode import OldCollapsedTree, OldCollapsedForest


def make_oldctree(tree):
    """Make an old CollapsedTree from an hDAG clade tree"""
    etetree = tree.to_ete(
        name_func=lambda n: n.attr["name"],
        features=["sequence"],
        feature_funcs={"abundance": lambda n: n.label.abundance},
    )
    for node in etetree.traverse():
        if not node.is_leaf():
            node.abundance = 0
    etetree.dist = 0
    for node in etetree.iter_descendants():
        node.dist = utils.hamming_distance(node.up.sequence, node.sequence)
    return OldCollapsedTree(etetree)


def make_oldcforest(newforest):
    """Make an old CollapsedForest"""
    return OldCollapsedForest(
        [make_oldctree(tree) for tree in newforest._forest.get_trees()]
    )


trees1 = pp.parse_outfile(
    "tests/example_output/original/small_outfile",
    abundance_file="tests/example_output/original/abundances.csv",
    root="GL",
)
# Sample trees disambiguated
trees1dis = [pp.disambiguate(tree.copy()) for tree in trees1]
trees2 = pp.parse_outfile(
    "tests/example_output/observed_root/small_outfile",
    abundance_file="tests/example_output/observed_root/abundances.csv",
    root="GL",
)
# Sample trees with observed root disambiguated
trees2dis = [pp.disambiguate(tree.copy()) for tree in trees2]

# Ambiguous and disambiguated, lists of sample trees and their associated counts
testtrees = [trees1, trees2]
testtreesdis = [trees1dis, trees2dis]

# The three kinds of CollapsedForests we're comparing:
# new ones with hDAG
newforests = [bp.CollapsedForest(trees) for trees in testtrees]

# new ones with a list of ctrees
newforests_ctrees = []
for forest in newforests:
    f = bp.CollapsedForest()
    f._ctrees = list(forest)
    f.n_trees = len(f._ctrees)
    f._cm_countlist = None
    newforests_ctrees.append(f)

# old ones
oldforests = [make_oldcforest(newforest) for newforest in newforests]

# all zipped
allforests = zip(newforests, newforests_ctrees, oldforests)

cmcount_dagfuncs = bp._cmcounter_dagfuncs()


def first(it):
    """Return the first element of an arbitrary iterable"""
    return list(zip(it, range(1)))[0][0]


def ll_isclose(ll1, ll2):
    """Check if likelihoods output by ll methods are the same"""
    return all(
        [
            np.isclose(ll1[0], ll2[0]),
            np.isclose(ll1[1][0], ll2[1][0]),
            np.isclose(ll1[1][1], ll2[1][1]),
        ]
    )


def test_newcounters():
    """Make sure the cmcounts found by new CollapsedTree init agree with old"""

    def pseudocount(mset):
        if (0, 1) in mset:
            return mset - {(0, 1)} + {(1, 1)}
        else:
            return mset

    for newforest, newforest_ctrees, oldforest in allforests:
        # To build cached _cm_counts
        newforest.ll(0.5, 0.5)
        newforest_ctrees.ll(0.5, 0.5)

        # Test just a single tree (the first in each forest)
        oldtreecounts = pseudocount(FrozenMultiset(oldforest.forest[0]._cm_list))
        newtreecounts = pseudocount(
            first(first(newforest._forest.get_trees()).weight_count(**cmcount_dagfuncs))
        )
        newtreectreecounts = FrozenMultiset(dict(first(newforest_ctrees)._cm_counts))

        assert oldtreecounts == newtreecounts
        assert oldtreecounts == newtreectreecounts

        # Test for the whole forests all together
        oldcmcounts = FrozenMultiset(
            [pseudocount(FrozenMultiset(ctree._cm_list)) for ctree in oldforest.forest]
        )

        newcmcounts_ctrees = FrozenMultiset()
        for treecounts, a in newforest_ctrees._cm_countlist:
            newcmcounts_ctrees += FrozenMultiset({FrozenMultiset(dict(treecounts)): a})

        newcmcounts = FrozenMultiset()
        for treecounts, a in newforest._cm_countlist:
            newcmcounts += FrozenMultiset({FrozenMultiset(dict(treecounts)): a})

        assert newcmcounts_ctrees == newcmcounts
        assert oldcmcounts == newcmcounts_ctrees


def test_newlikelihoods():
    """Make sure the likelihoods found by new CollapsedTree.ll agree with old"""
    p, q = 0.4, 0.6
    ll_dagfuncs = bp._ll_genotype_dagfuncs(p, q)

    # new and old trees agree
    for newforest, newforest_ctrees, oldforest in allforests:
        oldtreell = oldforest.forest[0].ll(p, q)
        newtreell = first(newforest).ll(p, q)
        newtreellctree = first(newforest_ctrees).ll(p, q)

        assert ll_isclose(oldtreell, newtreell)
        assert ll_isclose(oldtreell, newtreellctree)

        for marginal in [True, False]:
            oldfll = oldforest.ll(p, q, marginal=marginal)
            newfll = newforest.ll(p, q, marginal=marginal)
            newfctreell = newforest_ctrees.ll(p, q, marginal=marginal)

            assert ll_isclose(oldfll, newfll)
            assert ll_isclose(oldfll, newfctreell)
        for dagll, treell in zip(
            sorted(newforest._forest.weight_count(**ll_dagfuncs).elements()),
            sorted(ctree.ll(p, q) for ctree in newforest),
        ):
            assert np.isclose(dagll, treell)


def test_fit():
    """mle is the same from old code and computed directly using the DAG.
    This test extends ``gctree test`` (which verifies that inferred parameters
    on simulated forests consisting of a list of ctrees match simulation parameters)
    to also validate parameter inference in forests using the DAG to store trees.
    ``newforest_ctrees`` is the forest constructed from all the ctrees extracted from
    the DAG in ``newforest``"""
    for newforest, newforest_ctrees, oldforest in allforests:
        newforest._cm_countlist = None
        newforest_ctrees._cm_countlist = None
        assert newforest.n_trees == newforest_ctrees.n_trees
        pnew, qnew = newforest.mle(marginal=True)
        pnewc, qnewc = newforest_ctrees.mle(marginal=True)
        pold, qold = oldforest.mle(marginal=True)
        assert np.isclose(pnew, pold)
        assert np.isclose(qnew, qold)
        assert np.isclose(pnew, pnewc)
        assert np.isclose(qnew, qnewc)


def test_validate_ll_genotype():
    """compare old ll_genotype to new ll_genotype directly"""
    c_max, m_max = 10, 10
    for params in [(0.4, 0.6), (0.3, 0.5)]:
        for c in range(c_max):
            for m in range(m_max):
                if c > 0 or m > 1:
                    with np.errstate(all="raise"):
                        true_res = OldCollapsedTree._ll_genotype(c, m, *params)
                        res = bp.CollapsedTree._ll_genotype(c, m, *params)
                    assert np.isclose(true_res[0], res[0])
                    assert np.isclose(true_res[1][0], res[1][0])
                    assert np.isclose(true_res[1][1], res[1][1])


def test_recursion_depth():
    """Be sure ahead-of-time caching is implemented correctly to avoid
    recursion depth issues"""
    bp.CollapsedTree._ll_genotype.cache_clear()
    bp.CollapsedTree._max_ll_cache = {}
    with np.errstate(all="raise"):
        bp.CollapsedTree._ll_genotype(2, 500, 0.4, 0.6)


def test_context_likelihood():
    # These files will be present if pytest is run through `make test`.
    mutation_model = mm.MutationModel(
        mutability_file="HS5F_Mutability.csv", substitution_file="HS5F_Substitution.csv"
    )
    log_likelihood = mm._context_poisson_likelihood(mutation_model, splits=[])

    parent_seq = "AAGAAA"
    child_seq = "AATCAA"

    term1 = sum(
        log(
            mutation_model.mutability(fivemer)[0]
            * mutation_model.mutability(fivemer)[1][target_base]
        )
        for fivemer, target_base in [("AAGAA", "T"), ("AGAAA", "C")]
    )
    sum_mutabilities = sum(
        mutation_model.mutability(fivemer)[0]
        for fivemer in ["NNAAG", "NAAGA", "AAGAA", "AGAAA", "GAAAN", "AAANN"]
    )
    true_val = term1 + 2 * log(2 / sum_mutabilities) - 2
    assert true_val == log_likelihood(parent_seq, child_seq)

    # Now test chain split:
    parent_seq = parent_seq + parent_seq
    child_seq = child_seq + child_seq
    # At index 6, the second concatenated sequence starts.
    log_likelihood = mm._context_poisson_likelihood(mutation_model, splits=[6])

    true_val = 2 * term1 + 4 * log(4 / (2 * sum_mutabilities)) - 4
    assert true_val == log_likelihood(parent_seq, child_seq)



from gctree.isotyping import _isotype_dagfuncs, _isotype_annotation_dagfuncs
import gctree.branching_processes as bp
import gctree.isotyping as iso
import gctree.phylip_parse as pp
import ete3

testtrees = [
    ("((((A2)?,(G2,G2)?)?,G2,(A2,A2)?)?)M;", 3.0),
    ("((((A2)?,(G2,G2)?)?,A2,(A2,A2)?)?)M;", 4.0),
    ("((((A2)?,(G2,G3)?)?,G2,(A2,A2)?)?)M;", 5.0),
    ("((((A2)?,(G2,A2)?)?,G2,(A2,A2)?)?)M;", 4.0),
]

trees_seqcounts1 = pp.parse_outfile(
    "tests/example_output/original/small_outfile",
    abundance_file="tests/example_output/original/abundances.csv",
    root="GL",
)

forest = bp.CollapsedForest(trees_seqcounts1)
dag = forest._forest


def test_isotype_disambiguate():
    newisotype = iso.IsotypeTemplate(["M", "G3", "A1", "G2", "G4", "E", "A2"]).new
    for newick, weight in testtrees:
        tree = ete3.Tree(newick, format=8)
        for node in tree.traverse():
            node.isotype = newisotype(node.name)
        iso._disambiguate_isotype(tree)
        assert (
            sum(
                iso.isotype_distance(node.up.isotype, node.isotype)
                for node in tree.iter_descendants()
            )
            == weight
        )


def test_trim_byisotype():
    kwargs = _isotype_annotation_dagfuncs(
        isotypemap_file="example/isotypemap.txt",
        idmap_file="tests/example_output/original/idmap.txt",
    )
    # isotypemap_file=None,
    # idmap=None,
    # idmap_file=None,
    # isotype_names=None,
    tdag = dag.copy()
    tdag.optimal_weight_annotate(**kwargs, optimal_func=lambda l: l[0])  # noqa: E741
    for node in tdag.preorder():
        if node.attr is not None:
            node.attr["isotype"] = node._dp_data
    dag_filter = _isotype_dagfuncs()
    c = tdag.weight_count(**dag_filter)
    key = min(c)
    count = c[key]
    tdag.trim_optimal_weight(**dag_filter)
    assert tdag.weight_count(**dag_filter) == {key: count}
